#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import logging

from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.schema import Column, ForeignKey, Index, Table
from sqlalchemy.types import Integer, String

from Pegasus.db.admin.admin_loader import *
from Pegasus.db.admin.versions.base_version import BaseVersion
from Pegasus.db.schema import *
from Pegasus.db.schema import KeyInteger, check_table_exists, metadata, table_keywords

DB_VERSION = 6

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        """

        :param force:
        :return:
        """
        log.info("Updating to version %s" % DB_VERSION)
        # check whether it is safe to do that
        # self.db.execute("DROP TABLE file")

        # check if the migration was interrupted
        interrupted = False
        try:
            self.db.execute("SELECT site FROM rc_lfn_v4 LIMIT 0,1")
            interrupted = True
        except (OperationalError, ProgrammingError):
            pass

        # check if the migration is required
        try:
            if check_table_exists(self.db, RCLFN) and not interrupted:
                return
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            raise DBAdminError(e)

        # check if previous table exists. If not, the migration should not continue (for new dbs)
        try:
            if not interrupted:
                self.db.execute("SELECT site FROM rc_lfn LIMIT 0,1")
        except (OperationalError, ProgrammingError):
            return
        except Exception as e:
            raise DBAdminError(e)

        # Renaming rc_lfn.id to rc_lfn.lfn_id
        if interrupted:
            log.info("Recovering migration...")
            self._drop_table("rc_pfn")
            self._drop_table("rc_meta")
            self._drop_table("rc_lfn")
        else:
            log.info("Renaming 'rc_lfn' table...")
            if self.db.get_bind().driver == "mysqldb":
                self.db.execute("RENAME TABLE rc_lfn TO rc_lfn_v4")
            else:
                self.db.execute("ALTER TABLE rc_lfn RENAME TO rc_lfn_v4")

        # Create new rc_lfn table and populate it
        log.info("Updating tables...")
        self._create_table(RCLFN.__table__)
        self.db.commit()
        self._create_table(RCPFN.__table__)
        self._create_table(RCMeta.__table__)
        self.db.commit()

        try:
            log.info("Updating rc_lfn...")
            self.db.execute(
                "INSERT INTO rc_lfn(lfn) SELECT DISTINCT lfn FROM rc_lfn_v4"
            )
            self.db.commit()
            log.info("Updating rc_pfn...")
            self.db.execute(
                "INSERT INTO rc_pfn(lfn_id, pfn, site) SELECT l.lfn_id, a.pfn, a.site FROM rc_lfn l LEFT JOIN rc_lfn_v4 a ON (l.lfn=a.lfn)"
            )
            self.db.commit()
            log.info("Updating rc_meta...")
            self.db.execute(
                "INSERT INTO rc_meta(lfn_id, key, value) SELECT l.lfn_id, a.name, a.value FROM rc_lfn l LEFT JOIN rc_lfn_v4 b ON (l.lfn=b.lfn) INNER JOIN rc_attr a ON (a.id=b.id)"
            )
            self.db.commit()

        except Exception as e:
            log.info(e)
            self.db.rollback()

        # Drop old tables
        log.info("Cleaning up...")
        self._drop_table("rc_attr")
        self._drop_table("rc_lfn_v4")
        self.db.commit()

    def downgrade(self, force=False):
        """

        :param force:
        :return:
        """
        log.info("Downgrading from version %s" % DB_VERSION)

        log.info("Renaming 'rc_lfn' table...")
        if self.db.get_bind().driver == "mysqldb":
            self.db.execute("RENAME TABLE rc_lfn TO rc_lfn_v5")
        else:
            self.db.execute("ALTER TABLE rc_lfn RENAME TO rc_lfn_v5")
        self.db.commit()

        # Create tables
        log.info("Creating tables...")
        self._drop_index("v4_rc_lfn")
        self._drop_index("v4_rc_attr")
        metadata.remove(RCLFN.__table__)
        v4_rc_lfn = Table(
            "rc_lfn",
            metadata,
            Column("id", KeyInteger, primary_key=True, nullable=False),
            Column("lfn", String(245), nullable=False),
            Column("pfn", String(245), nullable=False),
            Column("site", String(245)),
            **table_keywords,
        )
        Index(
            "UNIQUE_RC_LFN",
            v4_rc_lfn.c.lfn,
            v4_rc_lfn.c.pfn,
            v4_rc_lfn.c.site,
            unique=True,
        )
        Index("v4_rc_lfn", v4_rc_lfn.c.lfn)
        v4_rc_lfn.create(self.db.get_bind(), checkfirst=True)

        v4_rc_attr = Table(
            "rc_attr",
            metadata,
            Column(
                "id",
                KeyInteger,
                ForeignKey("rc_lfn.id", ondelete="CASCADE"),
                primary_key=True,
                nullable=False,
            ),
            Column("name", String(245), primary_key=True, nullable=False),
            Column("value", String(245), nullable=False),
            **table_keywords,
        )
        Index("v4_rc_attr", v4_rc_attr.c.name)
        v4_rc_attr.create(self.db.get_bind(), checkfirst=True)
        v4_st_file = Table(
            "file",
            metadata,
            Column("file_id", KeyInteger, primary_key=True, nullable=False),
            Column(
                "task_id",
                KeyInteger,
                ForeignKey("task.task_id", ondelete="CASCADE"),
                nullable=True,
            ),
            Column("lfn", String(255), nullable=True),
            Column("estimated_size", Integer, nullable=True),
            Column("md_checksum", String(255), nullable=True),
            Column("type", String(255), nullable=True),
            **table_keywords,
        )

        Index("file_id_UNIQUE", v4_st_file.c.file_id, unique=True)
        Index("FK_FILE_TASK_ID", Task.task_id, unique=False)
        v4_st_file.create(self.db.get_bind(), checkfirst=True)

        # Migrate entries
        try:
            log.info("Updating rc_lfn...")
            self.db.execute(
                "INSERT INTO rc_lfn(lfn, pfn, site) SELECT l.lfn, a.pfn, a.site FROM (rc_lfn_v5 l INNER JOIN rc_pfn a ON (l.lfn_id=a.lfn_id))"
            )
            self.db.commit()
            log.info("Updating rc_attr...")
            self.db.execute(
                "INSERT INTO rc_attr(id, name, value) SELECT l.id, a.key, a.value FROM rc_lfn l LEFT JOIN rc_lfn_v5 b ON (l.lfn=b.lfn) INNER JOIN rc_meta a ON (a.lfn_id=b.lfn_id)"
            )
            self.db.commit()

        except Exception as e:
            log.info(e)
            self.db.rollback()

        # Drop tables
        log.info("Cleaning up...")
        self._drop_table("rc_pfn")
        self._drop_table("rc_meta")
        self._drop_table("workflow_files")
        self._drop_table("rc_lfn_v5")
        self.db.commit()

    def _create_table(self, table_obj):
        """
        Create a table from its backref.
        :param table_obj:
        :return:
        """
        try:
            table_obj.create(self.db.get_bind(), checkfirst=True)
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)

    def _drop_table(self, table_name):
        """
        Drop a table.
        :param table_name:
        :return:
        """
        try:
            self.db.execute("DROP TABLE %s" % table_name)
        except Exception:
            pass

    def _drop_index(self, index_name):
        """
        Drop am index.
        :param index_name:
        :return:
        """
        try:
            self.db.execute("DROP INDEX %s" % index_name)
        except Exception:
            pass



#!/usr/bin/env python3
import argparse
import logging
import sys
from pathlib import Path, PurePath

from cwl_utils import parser_v1_1 as cwl
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from Pegasus import yaml
from Pegasus.api import (
    Container,
    File,
    Job,
    ReplicaCatalog,
    Transformation,
    TransformationCatalog,
    Workflow,
)
from Pegasus.api.errors import DuplicateError

log = logging.getLogger("logger")

# --- logging ------------------------------------------------------------------
class ColoredFormatter(logging.Formatter):
    # printing debug level logs in yellow
    debug_format_colored = (
        "\u001b[33m%(asctime)s %(levelname)7s:  [%(funcName)s:%(lineno)d] "
        "%(message)s \u001b[0m"
    )

    debug_format = (
        "%(asctime)s %(levelname)7s:  [%(funcName)s:%(lineno)d] " "%(message)s"
    )

    def __init__(self):
        super().__init__("%(asctime)s %(levelname)7s:  %(message)s")

    def format(self, record):

        # save the original format
        fmt_orig = self._style._fmt

        if record.levelno == logging.DEBUG:
            self._style._fmt = ColoredFormatter.debug_format_colored
        else:
            self._style._fmt = ColoredFormatter.debug_format

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # revert format back to original
        self._style._fmt = fmt_orig

        return result


def setup_logger(debug_flag):

    # log to the console
    console = logging.StreamHandler()

    # default log level - make logger/console match
    log.setLevel(logging.INFO)
    console.setLevel(logging.INFO)

    # debug - from command line
    if debug_flag:
        log.setLevel(logging.DEBUG)
        console.setLevel(logging.DEBUG)

    # formatter
    console.setFormatter(ColoredFormatter())
    log.addHandler(console)
    log.debug("Logger has been configured")


# --- args ---------------------------------------------------------------------
def parse_args():
    parser = argparse.ArgumentParser(
        description=(
            "Converts a cwl workflow into Pegasus's native format.\n"
            "Note that this is a best-effort conversion and minor adjustments may\n"
            "be needed for proper execution."
        )
    )
    parser.add_argument(
        "cwl_workflow_file_path",
        help="Path to the file containing the CWL Workflow class.",
    )
    parser.add_argument(
        "workflow_inputs_file_path", help="YAML file describing the workflow inputs."
    )

    parser.add_argument(
        "transformation_spec_file_path",
        help=(
            "YAML file describing pegasus specific transformation properties.\n"
            "Given in the following format: {<tr name>: {site: <site name>, is_stageable: <bool>}, ...}"
        ),
    )

    parser.add_argument(
        "output_file_path", help="Desired path of the generated DAX file."
    )

    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        dest="debug",
        help="Enables debugging output.",
    )

    return parser.parse_args()


# --- cwl -> pegasus conversion functions --------------------------------------


def get_basename(name):
    return name.split("#")[1]


def get_name(parent_id: str, _id: str) -> str:
    return get_basename(parent_id) + "/" + get_basename(_id)


def load_wf_inputs(input_spec_file_path: str) -> dict:
    try:
        with open(input_spec_file_path) as f:
            wf_inputs = yaml.load(f)

        log.info("Loaded workflow inputs file: {}".format(input_spec_file_path))
    except FileNotFoundError:
        log.exception("Unable to find {}".format(input_spec_file_path))
        sys.exit(1)

    return wf_inputs


def load_tr_specs(tr_specs_file_path: str) -> dict:
    log.info("Validating {}".format(tr_specs_file_path))
    schema = {
        "type": "object",
        "patternProperties": {
            ".+": {
                "type": "object",
                "properties": {
                    "site": {"type": "string"},
                    "is_stageable": {"type": "boolean"},
                },
                "required": ["site", "is_stageable"],
                "additionalPropertes": False,
            }
        },
    }

    try:
        with open(tr_specs_file_path) as f:
            specs = yaml.load(f)

        validate(instance=specs, schema=schema)
    except ValidationError:
        log.exception(
            "Invalid transformation spec file. File should be in the following format:\n"
            "\t\t\t<tr name1>:\n"
            "\t\t\t    site: <site name>\n"
            "\t\t\t    is_stageable: <boolean>\n"
            "\t\t\t<tr name2>:\n"
            "\t\t\t    site: <site name>\n"
            "\t\t\t    is_stageable: <boolean>\n"
            "\t\t\t...\n"
        )
        sys.exit(1)
    except FileNotFoundError:
        log.exception(
            "Unable to find transformation spec file: {}".format(tr_specs_file_path)
        )
        sys.exit(1)

    log.info("Successfully loaded {}".format(tr_specs_file_path))

    return specs


def build_pegasus_tc(tr_specs: dict, cwl_wf: cwl.Workflow) -> TransformationCatalog:
    log.info("Building transformation catalog")
    tc = TransformationCatalog()

    for step in cwl_wf.steps:
        cwl_cmd_ln_tool = (
            cwl.load_document(step.run) if isinstance(step.run, str) else step.run
        )

        if cwl_cmd_ln_tool.baseCommand is None:
            raise ValueError("{} requires a 'baseCommand'".format(cwl_cmd_ln_tool.id))

        tool_path = PurePath(cwl_cmd_ln_tool.baseCommand)

        if not tool_path.is_absolute():
            raise ValueError(
                "{}.baseCommand: {} must be an absolute path".format(
                    cwl_cmd_ln_tool.id, cwl_cmd_ln_tool.baseCommand
                )
            )

        log.debug("baseCommand: {}".format(tool_path))

        # TODO: handle requirements (may not be needed or can manually add them in)
        site = "local"
        is_stageable = True

        try:
            site = tr_specs[tool_path.name]["site"]
            is_stageable = tr_specs[tool_path.name]["is_stageable"]
        except KeyError:
            log.warning(
                "Unable to look up transformation: {} in transformation spec file. Using defaults: site='local', is_stageable=True".format(
                    tool_path.name
                )
            )

        container_name = None
        if cwl_cmd_ln_tool.requirements:
            for req in cwl_cmd_ln_tool.requirements:
                if isinstance(req, cwl.DockerRequirement):
                    """
                    Currently not supported in DockerRequirement:
                    - dockerFile
                    - dockerImport
                    - dockerImageId
                    - dockerOutputDirectory
                    """

                    # assume to be docker because we can't distinguish between
                    # docker and singularity just by image name or url of zipped
                    # file
                    if req.dockerPull:
                        container_name = req.dockerPull

                        # planner won't allow deep lfn (for example: opensicencegrid/osg-el7 is invalid)
                        container_name = container_name.replace("/", "_")
                        image = "docker://" + req.dockerPull
                    elif req.dockerLoad:
                        image = req.dockerLoad
                        container_name = Path(req.dockerLoad).name
                    else:
                        raise NotImplementedError(
                            "Only DockerRequirement.dockerPull and DockerRequirement.dockerLoad currently supported"
                        )

                    try:
                        tc.add_containers(
                            Container(
                                container_name,
                                Container.DOCKER,
                                image,
                                image_site="local",
                            )
                        )
                    except DuplicateError:
                        pass

                    log.info(
                        "Added <Container name={}, container_type='docker', image={}, image_site='local'> from CommandLineTool: {}".format(
                            container_name, image, cwl_cmd_ln_tool.id
                        )
                    )
                    log.warning(
                        "Container types in the transformation catalog will need to be modified if containers are not of type: docker or if image file exists on a site other than 'local'"
                    )

                    #
                    break

        tr = Transformation(
            tool_path.name,
            site=site,
            pfn=str(tool_path),
            is_stageable=is_stageable,
            container=container_name,
        )
        log.debug(
            "tr = Transformation({}, site={}, pfn={}, is_stageable={})".format(
                tool_path.name, site, str(tool_path), is_stageable
            )
        )
        log.info("Adding <Transformation {}>".format(tr.name))

        try:
            tc.add_transformations(tr)
        except DuplicateError:
            log.warning(
                "<Transformation {}> is a duplicate and has already been added.".format(
                    tr.name
                )
            )

    log.info(
        "Building transformation catalog complete. {} transformations, {} containers added.".format(
            len(tc.transformations), len(tc.containers)
        )
    )

    return tc


def build_pegasus_rc(wf_inputs: dict, cwl_wf: cwl.Workflow) -> ReplicaCatalog:
    log.info("Building replica catalog")
    rc = ReplicaCatalog()

    for _input in cwl_wf.inputs:
        if _input.type == "File":
            input_name = get_basename(_input.id)

            try:
                current_wf_inputs = wf_inputs[input_name]
            except KeyError:
                log.exception(
                    "Unable to obtain input: {} from input spec file".format(input_name)
                )
                sys.exit(1)

            try:
                log.info(
                    "Adding replica: site={}, lfn={}, pfn={}".format(
                        "local", input_name, current_wf_inputs["path"]
                    )
                )
                # TODO: what about other sites?
                rc.add_replica("local", input_name, current_wf_inputs["path"])
            except KeyError:
                log.exception(
                    "Unable to obtain a path (pfn) for input file: {} from input spec file".format(
                        input_name
                    )
                )
                sys.exit(1)

    log.info(
        "Building replica catalog complete. {} entries added".format(len(rc.entries))
    )

    return rc


def collect_files(cwl_wf: cwl.Workflow) -> dict:
    log.info("Collecting workflow files")
    wf_files = dict()

    log.info("Parsing input files from workflow inputs")
    for _input in cwl_wf.inputs:
        if _input.type == "File":
            f = get_basename(_input.id)
            wf_files[f] = f
            log.info("Collected input file: {}".format(f))
            log.debug("wf_files[{0}] = {0}".format(f))

        elif (
            isinstance(_input.type, cwl.InputArraySchema)
            and _input.type.items == "File"
        ):

            raise NotImplementedError(
                "Support for File[] workflow input type in development"
            )

    log.info("Parsing output files from each workflow step")
    for step in cwl_wf.steps:
        cwl_cmd_ln_tool = (
            cwl.load_document(step.run) if isinstance(step.run, str) else step.run
        )

        for output in cwl_cmd_ln_tool.outputs:
            if output.type == "File":
                k = get_name(step.id, output.id)

                try:
                    v = output.outputBinding.glob
                except AttributeError:
                    v = None
                finally:
                    if not v:
                        raise ValueError(
                            "outputBinding.glob must be specified (e.g. file1.txt) for {}".format(
                                step.run
                            )
                        )

                if any(c in "*$" for c in v):
                    raise NotImplementedError(
                        "Unable to resolve wildcards in {}".format(v)
                    )

                wf_files[k] = v
                log.info("Collected output file: {}".format(k))
                log.debug("wf_files[{}] = {}".format(k, v))
            else:
                raise NotImplementedError(
                    "Support for output types other than File is in development"
                )

    log.info(
        "Collection of workflow files complete. {} files collected".format(
            len(wf_files)
        )
    )

    return wf_files


def collect_input_strings(wf_inputs: dict, cwl_wf: cwl.Workflow) -> dict:
    log.info("Collecting workflow input strings and string[] from input spec file")
    wf_input_str = dict()

    for _input in cwl_wf.inputs:
        if _input.type == "string" or (
            isinstance(_input.type, cwl.InputArraySchema)
            and _input.type.items == "string"
        ):
            input_name = get_basename(_input.id)

            try:
                wf_input_str[input_name] = wf_inputs[input_name]
                log.info("Collected input string: {}".format(input_name))
                log.debug(
                    "wf_input_str[{}] = {}".format(input_name, wf_inputs[input_name])
                )
            except KeyError:
                log.exception(
                    "Unable to obtain input: {} from input spec file".format(input_name)
                )
                sys.exit(1)

    log.info(
        "Collection of workflow input strings complete. {} string/string[] inputs collected".format(
            len(wf_input_str)
        )
    )

    return wf_input_str


def build_pegasus_wf(
    cwl_wf: cwl.Workflow, wf_files: dict, wf_input_str: dict
) -> Workflow:
    log.info("Building Pegasus workflow")

    wf = Workflow("cwl-converted-pegasus-workflow", infer_dependencies=True)

    for step in cwl_wf.steps:
        step_name = get_basename(step.id)
        log.info("Processing step: {}".format(step_name))
        cwl_cmd_ln_tool = (
            cwl.load_document(step.run) if isinstance(step.run, str) else step.run
        )

        job = Job(PurePath(cwl_cmd_ln_tool.baseCommand).name, _id=get_basename(step.id))

        # collect current step inputs
        log.info("Collecting step inputs from {}".format(step_name))
        step_inputs = dict()
        for _input in step.in_:
            input_id = get_basename(_input.id)

            step_inputs[input_id] = get_basename(_input.source)
            log.debug("step_inputs[{}] = {}".format(input_id, step_inputs[input_id]))

        # add inputs that are of type File
        for _input in cwl_cmd_ln_tool.inputs:
            if _input.type == "File":
                wf_file = File(wf_files[step_inputs[get_name(step.id, _input.id)]])

                job.add_inputs(wf_file)
                log.info("Step: {} added input file: {}".format(step_name, wf_file.lfn))
            """
            # TODO: handle File[] inputs
            elif isinstance(_input.type, cwl.CommandInputArraySchema):
                if _input.type.items == "File":
                    for f in step_inputs[get_name(step.id, _input.id)]:
                        wf_file = File(wf_files[f])

                        job.add_inputs(wf_file)
                        log.info(
                            "Step: {} added input file: {}".format(
                                step_name, wf_file.lfn
                            )
                        )
            """
        # add job outputs that are of type File
        log.info("Collecting step outputs from {}".format(step_name))
        for output in cwl_cmd_ln_tool.outputs:
            if output.type == "File":
                wf_file = File(wf_files[get_name(step.id, output.id)])

                job.add_outputs(wf_file)
                log.info(
                    "Step: {} added output file: {}".format(step_name, wf_file.lfn)
                )
            else:
                raise NotImplementedError(
                    "Support for output types other than File is in development"
                )

        # add job args
        args = (
            cwl_cmd_ln_tool.arguments
            if cwl_cmd_ln_tool.arguments is not None
            else list()
        )

        # args will be added in the order of their assigned inputBinding
        def get_input_binding(_input):
            key = 0
            if hasattr(_input, "inputBinding") and hasattr(
                _input.inputBinding, "position"
            ):
                key = _input.inputBinding.position

            return key if key else 0

        cwl_cmd_ln_tool_inputs = sorted(cwl_cmd_ln_tool.inputs, key=get_input_binding)

        for _input in cwl_cmd_ln_tool_inputs:
            # indicates whether or not input will appear in args
            if _input.inputBinding is not None:
                prefix = _input.inputBinding.prefix
                separate = _input.inputBinding.separate

                current_arg = ""
                if prefix:
                    current_arg += prefix

                if separate:
                    current_arg += " "

                if _input.type == "File":
                    current_arg += wf_files[step_inputs[get_name(step.id, _input.id)]]
                elif _input.type == "string":
                    current_arg += wf_input_str[
                        step_inputs[get_name(step.id, _input.id)]
                    ]

                # TODO: provide better support for array inputs being used in args (see https://www.commonwl.org/user_guide/09-array-inputs/index.html)
                elif isinstance(_input.type, cwl.CommandInputArraySchema):
                    separator = (
                        " "
                        if _input.inputBinding.itemSeparator is None
                        else _input.inputBinding.itemSeparator
                    )

                    if _input.type.items == "File":
                        current_arg += separator.join(
                            wf_files[f]
                            for f in step_inputs[get_name(step.id, _input.id)]
                        )
                    elif _input.type.items == "string":

                        current_arg += separator.join(
                            wf_input_str[step_inputs[get_name(step.id, _input.id)]]
                        )

                args.append(current_arg)

        job.add_args(*args)
        wf.add_jobs(job)

        log.info("Added job: {}".format(step.run))
        log.info("\tcmd: {}".format(job.transformation))
        log.info("\targs: {}".format(job.args))
        log.info("\tinputs: {}".format([f.lfn for f in job.get_inputs()]))
        log.info("\toutputs: {}".format([f.lfn for f in job.get_outputs()]))

    log.info("Building workflow complete. {} jobs added".format(len(wf.jobs)))

    return wf


def main():
    args = parse_args()
    setup_logger(args.debug)

    cwl_wf = cwl.load_document(args.cwl_workflow_file_path)
    log.info("Loaded cwl workflow: {}".format(args.cwl_workflow_file_path))

    wf_inputs = load_wf_inputs(args.workflow_inputs_file_path)
    tr_specs = load_tr_specs(args.transformation_spec_file_path)

    tc = build_pegasus_tc(tr_specs, cwl_wf)
    rc = build_pegasus_rc(wf_inputs, cwl_wf)

    wf_files = collect_files(cwl_wf)
    wf_input_strings = collect_input_strings(wf_inputs, cwl_wf)
    wf = build_pegasus_wf(cwl_wf, wf_files, wf_input_strings)

    wf.add_transformation_catalog(tc)
    wf.add_replica_catalog(rc)

    wf.write(file=args.output_file_path)
    log.info("Workflow written to {}".format(args.output_file_path))

    return 0


if __name__ == "__main__":
    sys.exit(main())



import errno
import os
import pwd
import shutil
import sys

from jinja2 import Environment, FileSystemLoader


class TutorialEnv:
    LOCAL_MACHINE = ("Local Machine Condor Pool", "submit-host")
    USC_HPCC_CLUSTER = ("USC HPCC Cluster", "usc-hpcc")
    OSG_FROM_ISI = ("OSG from ISI submit node", "osg")
    XSEDE_BOSCO = ("XSEDE, with Bosco", "xsede-bosco")
    BLUEWATERS_GLITE = ("Bluewaters, with Glite", "bw-glite")
    TACC_WRANGLER = ("TACC Wrangler with Glite", "wrangler-glite")
    OLCF_TITAN = ("OLCF TITAN with Glite", "titan-glite")
    OLCF_SUMMIT_KUBERNETES_BOSCO = (
        "OLCF Summit from Kubernetes using BOSCO",
        "summit-kub-bosco",
    )


class TutorialExample:
    PROCESS = ("Process", "process")
    PIPELINE = ("Pipeline", "pipeline")
    SPLIT = ("Split", "split")
    MERGE = ("Merge", "merge")
    EPA = ("EPA (requires R)", "r-epa")
    DIAMOND = ("Diamond", "diamond")
    CONTAINER = ("Population Modeling using Containers", "population")
    MPI = ("MPI Hello World", "mpi-hw")


def choice(question, options, default):
    "Ask the user to choose from a short list of named options"
    while True:
        sys.stdout.write("{} ({}) [{}]: ".format(question, "/".join(options), default))
        answer = sys.stdin.readline().strip()
        if len(answer) == 0:
            return default
        for opt in options:
            if answer == opt:
                return answer


def yesno(question, default="y"):
    "Ask the user a yes/no question"
    while True:
        sys.stdout.write("{} (y/n) [{}]: ".format(question, default))
        answer = sys.stdin.readline().strip().lower()
        if len(answer) == 0:
            answer = default
        if answer == "y":
            return True
        elif answer == "n":
            return False


def query(question, default=None):
    "Ask the user a question and return the response"
    while True:
        if default:
            sys.stdout.write("{} [{}]: ".format(question, default))
        else:
            sys.stdout.write("%s: " % question)
        answer = sys.stdin.readline().strip().replace(" ", "_")
        if answer == "":
            if default:
                return default
        else:
            return answer


def optionlist(question, options, default=0):
    "Ask the user to choose from a list of options"
    for i, option in enumerate(options):
        print("%d: %s" % (i + 1, option[0]))
    while True:
        sys.stdout.write("%s (1-%d) [%d]: " % (question, len(options), default + 1))
        answer = sys.stdin.readline().strip()
        if len(answer) == 0:
            return options[default][1]
        try:
            optno = int(answer)
            if optno > 0 and optno <= len(options):
                return options[optno - 1][1]
        except Exception:
            pass


class Workflow:
    def __init__(self, workflowdir, sharedir):
        self.jinja = Environment(loader=FileSystemLoader(sharedir), trim_blocks=True)
        self.name = os.path.basename(workflowdir)
        self.workflowdir = workflowdir
        self.sharedir = sharedir
        self.properties = {}
        self.home = os.environ["HOME"]
        self.user = pwd.getpwuid(os.getuid())[0]
        self.tutorial = None
        self.generate_tutorial = False
        self.tutorial_setup = None
        self.compute_queue = "default"
        self.project = "MYPROJ123"
        sysname, _, _, _, machine = os.uname()
        if sysname == "Darwin":
            self.os = "MACOSX"
        else:
            # Probably Linux
            self.os = sysname.upper()
        self.arch = machine

    def copy_template(self, template, dest, mode=0o644):
        "Copy template to dest in workflowdir with mode"
        path = os.path.join(self.workflowdir, dest)
        t = self.jinja.get_template(template)
        t.stream(**self.__dict__).dump(path)
        os.chmod(path, mode)

    def copy_dir(self, src, dest):
        # self.mkdir(dest)
        if not src.startswith("/"):
            src = os.path.join(self.sharedir, src)
        try:
            dest = os.path.join(self.workflowdir, dest)
            shutil.copytree(src, dest)
        except OSError as exc:  # python >2.5
            if exc.errno == errno.ENOTDIR:
                shutil.copy(src, dest)
            else:
                raise

    def mkdir(self, path):
        "Make relative directory in workflowdir"
        path = os.path.join(self.workflowdir, path)
        if not os.path.exists(path):
            os.makedirs(path)

    def configure(self):
        # The tutorial is a special case
        if yesno("Do you want to generate a tutorial workflow?", "n"):
            self.config = "tutorial"
            self.daxgen = "tutorial"
            self.generate_tutorial = True

            # determine the environment to setup tutorial for
            self.tutorial_setup = optionlist(
                "What environment is tutorial to be setup for?",
                [
                    TutorialEnv.LOCAL_MACHINE,
                    TutorialEnv.USC_HPCC_CLUSTER,
                    TutorialEnv.OSG_FROM_ISI,
                    TutorialEnv.XSEDE_BOSCO,
                    TutorialEnv.BLUEWATERS_GLITE,
                    TutorialEnv.TACC_WRANGLER,
                    TutorialEnv.OLCF_TITAN,
                    TutorialEnv.OLCF_SUMMIT_KUBERNETES_BOSCO,
                ],
            )

            # figure out what example options to provide
            examples = [
                TutorialExample.PROCESS,
                TutorialExample.PIPELINE,
                TutorialExample.SPLIT,
                TutorialExample.MERGE,
                TutorialExample.EPA,
                TutorialExample.CONTAINER,
            ]
            if self.tutorial_setup != "osg":
                examples.append(TutorialExample.DIAMOND)

            if self.tutorial_setup in [
                "bw-glite",
                "wrangler-glite",
                "titan-glite",
                "summit-kub-bosco",
            ]:
                examples.append(TutorialExample.MPI)
                self.project = query(
                    "What project your jobs should run under. For example on TACC there are like : TG-DDM160003 ?"
                )

            self.tutorial = optionlist("What tutorial workflow do you want?", examples)

            self.setup_tutorial()
            return

        # Determine which DAX generator API to use
        self.daxgen = choice(
            "What DAX generator API do you want to use?",
            ["python", "perl", "java", "r"],
            "python",
        )

        # Determine what kind of site catalog we need to generate
        self.config = optionlist(
            "What does your computing infrastructure look like?",
            [
                ("Local Machine Condor Pool", "condorpool"),
                ("Remote Cluster using Globus GRAM", "globus"),
                ("Remote Cluster using CREAMCE", "creamce"),
                ("Local PBS Cluster with Glite", "glite"),
                ("Remote PBS Cluster with BOSCO and SSH", "bosco"),
            ],
        )

        # Find out some information about the site
        self.sitename = query("What do you want to call your compute site?", "compute")
        self.os = choice(
            "What OS does your compute site have?", ["LINUX", "MACOSX"], self.os
        )
        self.arch = choice(
            "What architecture does your compute site have?",
            ["x86_64", "x86"],
            self.arch,
        )

    def setup_tutorial(self):
        """
        Set up tutorial for pre-defined computing environments
        :return:
        """

        if self.tutorial_setup is None:
            self.tutorial_setup = "submit-host"

        if self.tutorial_setup == "submit-host":
            self.sitename = "condorpool"
        elif self.tutorial_setup == "usc-hpcc":
            self.sitename = "usc-hpcc"
            self.config = "glite"
            self.compute_queue = "quick"
            # for running the whole workflow as mpi job
            self.properties["pegasus.job.aggregator"] = "mpiexec"
        elif self.tutorial_setup == "osg":
            self.sitename = "osg"
            self.os = "linux"
            if not yesno("Do you want to use Condor file transfers", "y"):
                self.staging_site = "isi_workflow"
        elif self.tutorial_setup == "xsede-bosco":
            self.sitename = "condorpool"
        elif self.tutorial_setup == "bw-glite":
            self.sitename = "bluewaters"
            self.config = "glite"
            self.compute_queue = "normal"
        elif self.tutorial_setup == "wrangler-glite":
            self.sitename = "wrangler"
            self.config = "glite"
            self.compute_queue = "normal"
        elif self.tutorial_setup == "titan-glite":
            self.sitename = "titan"
            self.config = "glite"
            self.compute_queue = "titan"
        elif self.tutorial_setup == "summit-kub-bosco":
            self.sitename = "summit"
            self.config = "bosco"
            self.compute_queue = "batch"
        return

    def generate(self):
        os.makedirs(self.workflowdir)
        if self.tutorial != "population":
            self.mkdir("input")
        self.mkdir("output")

        if self.generate_tutorial:
            self.copy_template("%s/tc.txt" % self.tutorial, "tc.txt")

            if self.tutorial == "r-epa":
                self.copy_template("%s/daxgen.R" % self.tutorial, "daxgen.R")
            elif self.tutorial != "mpi-hw":
                self.copy_template("%s/daxgen.py" % self.tutorial, "daxgen.py")

            if self.tutorial == "diamond":

                # Executables used by the diamond workflow
                self.mkdir("bin")
                self.copy_template(
                    "diamond/transformation.py", "bin/preprocess", mode=0o755
                )
                self.copy_template(
                    "diamond/transformation.py", "bin/findrange", mode=0o755
                )
                self.copy_template(
                    "diamond/transformation.py", "bin/analyze", mode=0o755
                )

                # Diamond input file
                self.copy_template("diamond/f.a", "input/f.a")
            elif self.tutorial == "split":
                # Split workflow input file
                self.mkdir("bin")
                self.copy_template("split/pegasus.html", "input/pegasus.html")
            elif self.tutorial == "r-epa":
                # Executables used by the R-EPA workflow
                self.mkdir("bin")
                self.copy_template(
                    "r-epa/epa-wrapper.sh", "bin/epa-wrapper.sh", mode=0o755
                )
                self.copy_template("r-epa/setupvar.R", "bin/setupvar.R", mode=0o755)
                self.copy_template(
                    "r-epa/weighted.average.R", "bin/weighted.average.R", mode=0o755
                )
                self.copy_template(
                    "r-epa/cumulative.percentiles.R",
                    "bin/cumulative.percentiles.R",
                    mode=0o755,
                )
            elif self.tutorial == "population":
                self.copy_template("%s/Dockerfile" % self.tutorial, "Dockerfile")
                self.copy_template("%s/Singularity" % self.tutorial, "Singularity")
                self.copy_template(
                    "%s/tc.txt.containers" % self.tutorial, "tc.txt.containers"
                )
                self.copy_dir("%s/scripts" % self.tutorial, "scripts")
                self.copy_dir("%s/data" % self.tutorial, "input")
                # copy the mpi wrapper, c code and mpi
            elif self.tutorial == "mpi-hw":
                # copy the mpi wrapper, c code and mpi example
                # Executables used by the mpi-hw workflow
                self.mkdir("bin")
                self.copy_template(
                    "%s/pegasus-mpi-hw.c" % self.tutorial, "pegasus-mpi-hw.c"
                )
                self.copy_template("%s/Makefile" % self.tutorial, "Makefile")
                self.copy_template("%s/daxgen.py.template" % self.tutorial, "daxgen.py")
                self.copy_template(
                    "%s/mpi-hello-world-wrapper" % self.tutorial,
                    "bin/mpi-hello-world-wrapper",
                    mode=0o755,
                )
                self.copy_template("split/pegasus.html", "input/f.in")

        else:
            self.copy_template("tc.txt", "tc.txt")
            if self.daxgen == "python":
                self.copy_template("daxgen/daxgen.py", "daxgen.py")
            elif self.daxgen == "perl":
                self.copy_template("daxgen/daxgen.pl", "daxgen.pl")
            elif self.daxgen == "java":
                self.copy_template("daxgen/DAXGen.java", "DAXGen.java")
            elif self.daxgen == "r":
                self.copy_template("daxgen/daxgen.R", "daxgen.R")
            else:
                assert False

        self.copy_template("sites.xml", "sites.xml")
        self.copy_template("plan_dax.sh", "plan_dax.sh", mode=0o755)
        self.copy_template("plan_cluster_dax.sh", "plan_cluster_dax.sh", mode=0o755)
        self.copy_template("generate_dax.sh", "generate_dax.sh", mode=0o755)
        self.copy_template("README.md", "README.md")
        self.copy_template("rc.txt", "rc.txt")
        self.copy_template("pegasus.properties", "pegasus.properties")

        if self.tutorial == "diamond":
            if self.tutorial_setup == "wrangler-glite":
                self.copy_template(
                    "pmc-wrapper.wrangler", "bin/pmc-wrapper", mode=0o755
                )
            elif self.tutorial_setup == "titan-glite":
                self.copy_template("pmc-wrapper.titan", "bin/pmc-wrapper", mode=0o755)
            elif self.tutorial_setup == "wrangler-glite":
                self.copy_template(
                    "pmc-wrapper.wrangler", "bin/pmc-wrapper", mode=0o755
                )
            elif self.tutorial_setup == "summit-kub-bosco":
                self.copy_template("pmc-wrapper.summit", "bin/pmc-wrapper", mode=0o755)

        if self.generate_tutorial:
            sys.stdout.write(
                "Pegasus Tutorial setup for example workflow - %s for execution on %s in directory %s\n"
                % (self.tutorial, self.tutorial_setup, self.workflowdir)
            )


def usage():
    print("Usage: %s WORKFLOW_DIR" % sys.argv[0])


def main(pegasus_share_dir):
    if len(sys.argv) != 2:
        usage()
        exit(1)

    if "-h" in sys.argv:
        usage()
        exit(1)

    workflowdir = sys.argv[1]
    if os.path.exists(workflowdir):
        print("ERROR: WORKFLOW_DIR '%s' already exists" % workflowdir)
        exit(1)

    workflowdir = os.path.abspath(workflowdir)
    sharedir = os.path.join(pegasus_share_dir, "init")
    w = Workflow(workflowdir, sharedir)
    w.configure()
    w.generate()



"""
Base classes that define the interface that the back-end specific
retrieval classes must expose.  Each read-only property includes
inline epydoc describing what is to be returned and what type
(string, int, an object) each property must return when called.

How a subclass caches attributes or performs lazy evaluation is
implementation-specific and not defined here.  The only conformity
necessary is that the proper property returns the correct data as
the correct type.

Any class attributes or methods that are specific to the subclass
implementations MUST start with an underscore:

ie: self._wf_uuid = None or def _calculateValue(self, job):

This signals that the attr or method is subclass-specific and
also allows the inherited __repr__() method to ignore it.
"""

__author__ = "Monte Goode MMGoode@lbl.gov"

import logging


class WorkflowBase:
    # indent level for pretty printing = override in subclasses
    # if you want different indent levels for your various
    # objects.
    _indent = 1

    def __init__(self):
        self.log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )

    def __repr__(self):
        spacer = "  "
        retval = "%s:" % self.__class__
        if self._indent > 1:
            retval = "\n{}+++ {}:".format(spacer * self._indent, self.__class__)
        for i in dir(self):
            if (
                i.startswith("_")
                or i == "initialize"
                or i == "db"
                or i == "metadata"
                or i == "session"
                or i == "log"
            ):
                continue
            try:
                retval += "\n{}* {} : {}".format(
                    spacer * self._indent, i, eval("self.%s" % i),
                )
            except NotImplementedError as e:
                retval += "\n{}* {} : WARNING: {}".format(spacer * self._indent, i, e)
        return retval


class Workflow(WorkflowBase):
    """
    Top level workflow class that exposes information about
    a specific workflow and the associated jobs/etc.

    Usage::

     w = Workflow()
     w.initialize('unique_wf_uuid')
     print w.timestamp, w.dax_label

    etc
    """

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, wf_id):
        """
        This method is the initialization method that accepts
        the unique wf_uuid and triggers the subclass specific
        queries and calculations that pulls the workflow
        information from the back end.

        The wf_id is represented in the .bp logs as "wf.id".

        @type   wf_id: string
        @param  wf_id: the unique wf_uuid as defined by Pegasus
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def wf_uuid(self):
        """
        Return the wf_uuid for this workflow.

        @rtype:     string
        @return:    The wf_uuid of the current workflow
        """
        raise NotImplementedError("wf_uuid not yet implemented")

    @property
    def dax_label(self):
        """
        Return dax_label from storage backend.

        @rtype:     string
        @return:    The dax_label of the current workflow.
        """
        raise NotImplementedError("dax_label not yet implemented")

    @property
    def timestamp(self):
        """
        Return timestamp from storage backend.

        @rtype:     python datetime obj (utc)
        @return:    The workflow timestamp
        """
        raise NotImplementedError("timestamp not yet implemented")

    @property
    def submit_hostname(self):
        """
        Return submit_hostname from storage backend.

        @rtype:     string
        @return:    The workflow submit host
        """
        raise NotImplementedError("submit_hostname not yet implemented")

    @property
    def submit_dir(self):
        """
        Return submid_dir from storage backend.

        @rtype:     string
        @return:    The workflow submit directory
        """
        raise NotImplementedError("submit_dir not yet implemented")

    @property
    def planner_arguments(self):
        """
        Return planner_arguments from storage backend.

        @rtype:     string
        @return:    The workflow planner arguments
        """
        raise NotImplementedError("planner_arguments not yet implemented")

    @property
    def user(self):
        """
        Return user from storage backend.

        @rtype:     string
        @return:    The workflow user
        """
        raise NotImplementedError("user not yet implemented")

    @property
    def grid_dn(self):
        """
        Return grid_dn from storage backend.

        @rtype:     string
        @return:    The grid DN of the workflow
        """
        raise NotImplementedError("grid_dn not yet implemented")

    @property
    def planner_version(self):
        """
        Return planner_version from storage backend.

        @rtype:     string
        @return:    The planner version of the workflow
        """
        raise NotImplementedError("planner_version not yet implemented")

    @property
    def parent_wf_uuid(self):
        """
        Return parent_wf_uuid from storage backend.

        @rtype:     string
        @return:    The parent wf_uuid if it exists
        """
        raise NotImplementedError("parent_wf_uuid not yet implemented")

    @property
    def sub_wf_uuids(self):
        """
        Returns a list of the wf_uuids of any sub-workflows associated
        with the current workflow object.  Returned in the order in
        which they are entered in the workflow table.  If no sub-workflows
        are found, return an empty list.

        @rtype:     List of strings
        @return:    The wf_uuids of any sub-workflows.
        """
        raise NotImplementedError("sub_wf_uuids not yet implemented")

    @property
    def start_events(self):
        """
        Return a list of Workflowstate object instances representing
        the re/start events.  The list should ordered by timestamp.

        In the event that there are no logged workflow states an empty
        list should be returned.

        In the case that there is a dropped event (ie: no matching end
        event to a start event or vice versa), the missing event will
        be padded as a None.  This is an error situation.

        @rtype:     List of Workflowstate object instances (or None)
        @return:    Returns a list with workflow start events.
        """
        raise NotImplementedError("start_events not yet implemented")

    @property
    def end_events(self):
        """
        Return a list of Workflowstate object instances representing
        the end events.  The list should ordered by timestamp.

        In the event that there are no logged workflow states an empty
        list should be returned.

        In the case that there is a dropped event (ie: no matching end
        event to a start event or vice versa), the missing event will
        be padded as a None.  This is an error situation.

        @rtype:     List of Workflowstate object instances (or None)
        @return:    Returns a list with workflow end events.
        """
        raise NotImplementedError("end_events not yet implemented")

    @property
    def is_running(self):
        """
        Derived boolean flag indicating if the workflow
        is currently running.  Derived in a backend-appropriate
        way.

        @rtype:     boolean
        @return:    Indicates if the workflow is running.
        """
        raise NotImplementedError("is_running not yet implemented")

    @property
    def is_restarted(self):
        """
        Derived boolean flag indicating if this workflow has
        been retstarted.  Derived in a backend-appropriate
        way.

        @rtype:     boolean
        @return:    Indicates if the workflow has been restarted.
        """
        raise NotImplementedError("is_restarted not yet implemented")

    @property
    def restart_count(self):
        """
        Returns an integer reflecting restart count.  Derived in
        a backend-appropriate way.

        @rtype:     integer
        @return:    Number of workflow restarts.
        """
        raise NotImplementedError("restart_count not yet implemented")

    @property
    def total_time(self):
        """
        Returns the total runtime of the workflow.  This is defined
        as the delta between either the first start event and last end
        event, or if the job is still running, between the first
        start event and current epoch UTC time.

        @rtype:     python datetime.timedelta object or None
        @return:    The total time of the workflow.
        """
        raise NotImplementedError("total_time not yet implemented")

    @property
    def jobs(self):
        """
        Returns a list of the jobs associated with this workflow
        object.  This property is a prime candidate for lazy eval
        as there is no need to query the backend for this information
        if the user is only looking at superficial information about
        the workflow - ie: if it is running, how long it took, etc.

        @rtype:     list of Job objects
        @return:    List of job objects associated with current wf
        """
        raise NotImplementedError("jobs not yet implemented")

    @property
    def total_jobs_executed(self):
        """
        Return the number of jobs that were executed as an
        integer value.

        @rtype:     integer
        @return:    Number of jobs executed
        """
        raise NotImplementedError("total_jobs_executed not yet implemented")

    @property
    def successful_jobs(self):
        """
        Return the number of jobs that executed successfully
        as an integer value.

        @rtype:     integer
        @return:    Number of sucessfully executed jobs
        """
        raise NotImplementedError("successful_jobs not yet implemented")

    @property
    def failed_jobs(self):
        """
        Return the number of jobs that failed as an integer
        value.

        @rtype:     integer
        @return:    Number of failed jobs
        """
        raise NotImplementedError("failed_jobs not yet implemented")

    @property
    def restarted_jobs(self):
        """
        Return the number of jobs that were restarted.

        @rtype:     integer
        @return:    Number of restarted jobs
        """
        raise NotImplementedError("restarted_jobs not yet implemented")

    @property
    def submitted_jobs(self):
        """
        Return the number of jobs that were submitted.

        @rtype:     integer
        @return:    Number of submitted jobs
        """
        raise NotImplementedError("submitted_jobs not yet implemented")

    @property
    def jobtypes_executed(self):
        """
        Returns a dictionary of the various jobtypes that
        are executed in the current workflow and a count of
        how many of each type.

        Example: {'create dir': 1, 'compute': 105}

        @rtype:     dict - string keys, integer values
        @return:    A dictionary of a count of the jobtypes
                    that were executed in the current workflow.
        """
        raise NotImplementedError("jobtypes_executed not yet implemented")


class Workflowstate(WorkflowBase):
    """
    Class to expose information about a specific
    workflow event.  This is a simple class to expose state
    and timestamp information as class attributes.

    Usage::

     ws = Workflowstate()
     ws.initialize(state, timestamp)
     print ws.state

    etc
    """

    _indent = 2

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, state, timestamp):
        """
        This method is the initialization method that accepts
        the state and timestamp of a given workflow state
        event.

        @type   state: string
        @param  state: the jobstate entry as defined by Pegasus.
        @type   timestamp: float
        @param  timestamp: the epoch timestamp as reported by Pegasus.
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def state(self):
        """
        Return the current workflowstate state.  Might be
        none if there is no state information logged yet.

        @rtype:     string or None
        @return:    Return current job state
        """
        raise NotImplementedError("state not implemented yet")

    @property
    def timestamp(self):
        """
        Return the timestamp of the current workflow state.  Might be
        none if there is no state information logged yet.

        @rtype:     python datetime obj (utc) or None
        @return:    Return timestamp of current job state
        """
        raise NotImplementedError("timestamp not implemented yet")


class Job(WorkflowBase):
    """
    Class to retrieve and expose information about a
    specific job.  This class is intended to be instantiated
    inside a Workflow() object and not as a stand-alone
    instance.

    Usage::

     j = Job()
     j.initialize('unique_wf_uuid', 3)
     print j.name

    etc
    """

    _indent = 2

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, wf_id, job_id):
        """
        This method is the initialization method that accepts
        the unique wf_uuid and and job_submit_seq that triggers
        the subclass specific queries and calculations that
        pulls job information from the back end.

        The wf_id is represented in the .bp logs as "wf.id".
        The job_id is represented as "job.id".

        @type   wf_id: string
        @param  wf_id: the unique wf_uuid as defined by Pegasus
        @type   job_id: integer
        @param  job_id: the sequence number as defined
                by tailstatd
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def job_submit_seq(self):
        """
        Return job_submit_seq of current job (an input arg).

        @rtype:     integer
        @return:    Return job_submit_seq of current job
        """
        raise NotImplementedError("job_submit_seq not yet implemented")

    @property
    def name(self):
        """
        Return the job name from the storage backend.

        @rtype:     string
        @return:    Return job name
        """
        raise NotImplementedError("name not yet implemented")

    @property
    def host(self):
        """
        Return job host information from storage backend.

        @rtype:     Host object instance
        @return:    Return a host object with host info for
                    current job.
        """
        raise NotImplementedError("host not yet implemented")

    @property
    def condor_id(self):
        """
        Return the condor_id from the storage backend.

        @rtype:     string (looks like a float however)
        @return:    Return job condor_id
        """
        raise NotImplementedError("condor_id not yet implemented")

    @property
    def jobtype(self):
        """
        Return jobtype from the storage backend.

        @rtype:     string
        @return:    Return jobtype
        """
        raise NotImplementedError("jobtype not yet implemented")

    @property
    def clustered(self):
        """
        Return the clustered boolean flag from the storage
        backend.  This may need to be derived depending on
        how the backend implementation does/not store this
        value.

        @rtype:     boolean
        @return:    Return True or False depending on if the
                    job is clustered or not.
        """
        raise NotImplementedError("clustered not yet implemented")

    @property
    def site_name(self):
        """
        Return the site name from the storage backend.

        @rtype:     string
        @return:    Return site_name for current job
        """
        raise NotImplementedError("site_name not yet implemented")

    @property
    def remote_user(self):
        """
        Return the remote use of the current job from
        the storage backend.

        @rtype:     string
        @return:    Return remote_user for current job.
        """
        raise NotImplementedError("remote_user not yet implemented")

    @property
    def remote_working_dir(self):
        """
        Return the remote working directory of the current
        job from the storage backend.

        @rtype:     string
        @return:
        """
        raise NotImplementedError("remote_working_dir not yet implemented")

    @property
    def cluster_start_time(self):
        """
        Return the job cluster start time as a python
        datetime object (utc) if it exists or None
        if it does not.  Not all jobs will have this
        value.

        @rtype:     python datetime obj (utc) or None
        @return:    Return job cluster start time.
        """
        raise NotImplementedError("cluster_start_time not yet implemented")

    @property
    def cluster_duration(self):
        """
        Return the job cluster duration from the
        storage backend as a float or None if this value
        is not assocaited with the current job.  Not all j
        will have this value.

        @rtype:     float (from db)
        @return:
        """
        raise NotImplementedError("cluster_duration not yet implemented")

    @property
    def tasks(self):
        """
        Returns a list of the tasks associated with this job
        object.  This property is a prime candidate for lazy eval
        as there is no need to query the backend for this information
        if the user is only looking at superficial information about
        the job - ie: its current state, name, etc.

        @rtype:     list of Task objects
        @return:    List of task objects associated with current job
        """
        raise NotImplementedError("tasks not yet implemented")

    @property
    def is_restart(self):
        """
        Return a boolean flag indicating whether or not this
        curent job is a "restart".  This value will be derived
        from backend information as appropriate.

        @rtype:     boolean
        @return:    Return True or False if the job is a restart
                    or not.
        """
        raise NotImplementedError("is_restart not yet implemented")

    @property
    def is_success(self):
        """
        Return a boolean flag indicating whether or not this
        curent job was successful.  This value will be derived
        from backend information as appropriate.

        @rtype:     boolean
        @return:    Return True or False if the job is a restart
        """
        raise NotImplementedError("is_success not yet implemented")

    @property
    def is_failure(self):
        """
        Return a boolean flag indicating whether or not this
        curent job has failed.  This value will be derived
        from backend information as appropriate.

        @rtype:     boolean
        @return:    Return True or False if the job is a restart
        """
        raise NotImplementedError("is_failure not yet implemented")

    @property
    def current_state(self):
        """
        Return the current state of this job.  This property
        pretty much requires lazy evaluation every access rather
        than attribute caching.  A single job moves through
        multiple jobstates and this property should return
        the current state of the running job when accessed.

        In the event that there is not yet a jobstate logged for
        this job, the Jobstate instance will have its properties
        "state" and "timestamp" set to None.

        @rtype:     Jobstate object instance
        @return:    Return the current/last jobstate event.
        """
        raise NotImplementedError("current_state not yet implemented")

    @property
    def all_jobstates(self):
        """
        Return a list of Jobstate object instances representing
        the states that the job has moved through.  The list should
        ordered by the order of the different jobstate submissions.

        In the event that there are no logged jobstates an empty
        list should be returned.

        This property may do light weight attribute caching, but
        the current jobstate should still be lazily evaluated
        and the list updated if need be.

        @rtype:     List of Jobstate object instances
        @return:    Returns a list with all the jobstates this job
                    has moved through.
        """
        raise NotImplementedError("all_jobstates not yet implemented")

    @property
    def submit_time(self):
        """
        Return the timestamp of when this job was submitted.

        @rtype:     python datetime obj (utc) or None
        @return:    Return the submit time of this job
        """
        raise NotImplementedError("submit_time not yet implemented")

    @property
    def elapsed_time(self):
        """
        Return the elapsed time of this job.  Calculated as
        the delta between the submit time and the current/last
        jobstate timestamp.

        @rtype:     python datetime.timedelta or None
        @return:    Return the elapsed time of this job
        """
        raise NotImplementedError("elapsed_time not yet implemented")

    @property
    def edge_parents(self):
        """
        Return a list of job objects for the parent job edges for
        this current job object.  The job objects returned by this
        property will NOT contain additional edge information (ie: this
        method will return an empty list) to avoid a recursive situation.

        @rtype:     list containing Job objects
        @return:    Return the parent edge Job objects.
        """
        raise NotImplementedError("edge_parents not yet implemented")

    @property
    def edge_children(self):
        """
        Return a list of job objects for the child job edges for
        this current job object.  The job objects returned by this
        property will NOT contain additional edge information (ie: this
        method will return an empty list) to avoid a recursive situation.

        @rtype:     list containing Job objects
        @return:    Return the child edge Job objects.
        """
        raise NotImplementedError("edge_children not yet implemented")


class Jobstate(WorkflowBase):
    """
    A small class that returns jobstate information.  Intended
    to be instantiated by a call to job.current_state.  Is
    not cached so multiple calls will return the latest information.

    Usage::

     js = Jobstate()
     js.initialize('unique_wf_id', 3)
     print js.state, js.timestamp

    etc
    """

    _indent = 3

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, wf_id, job_id):
        """
        This method is the initialization method that accepts
        the unique wf_uuid and and job_submit_seq that triggers
        the subclass specific queries and calculations that
        pulls host information from the back end.

        The wf_id is represented in the .bp logs as "wf.id".
        The job_id is represented as "job.id".

        @type   wf_id: string
        @param  wf_id: the unique wf_uuid as defined by Pegasus
        @type   job_id: integer
        @param  job_id: the sequence number as defined
                by tailstatd
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def state(self):
        """
        Return the current jobstate state.  Might be
        none if there is no state information logged yet.

        @rtype:     string or None
        @return:    Return current job state
        """
        raise NotImplementedError("state not implemented yet")

    @property
    def timestamp(self):
        """
        Return the timestampe of the current job state.  Might be
        none if there is no state information logged yet.

        @rtype:     python datetime obj (utc) or None
        @return:    Return timestamp of current job state
        """
        raise NotImplementedError("timestamp not implemented yet")


class Host(WorkflowBase):
    """
    A straightforward class that contains host information
    about a job.  This is intended to be instantiated inside
    a Job() object and not as a standalone instance.

    Usage::

     h = Host()
     h.initialize('unique_wf_uuid', 3)
     print h.site_name, h.hostname

    etc.
    """

    _indent = 3

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, wf_id, job_id):
        """
        This method is the initialization method that accepts
        the unique wf_uuid and and job_submit_seq that triggers
        the subclass specific queries and calculations that
        pulls host information from the back end.

        The wf_id is represented in the .bp logs as "wf.id".
        The job_id is represented as "job.id".

        @type   wf_id: string
        @param  wf_id: the unique wf_uuid as defined by Pegasus
        @type   job_id: integer
        @param  job_id: the sequence number as defined
                by tailstatd
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def site_name(self):
        """
        Return the site name associated with this host.
        Might be None if a host has not been associated
        with a particular job at time of calling.

        @rtype:     string or None
        @return:    Return host site name or None
        """
        raise NotImplementedError("site_name not yet implemented")

    @property
    def hostname(self):
        """
        Return the host name associated with this host.
        Might be None if a host has not been associated
        with a particular job at time of calling.

        @rtype:     string or None
        @return:    Return hostname or None
        """
        raise NotImplementedError("hostname not yet implemented")

    @property
    def ip_address(self):
        """
        Return the ip address associated with this host.
        Might be None if a host has not been associated
        with a particular job at time of calling.

        @rtype:     string or None
        @return:    Return host ip address or None
        """
        raise NotImplementedError("ip_address not yet implemented")

    @property
    def uname(self):
        """
        Return the uname information of this host machine.
        Might be None if a host has not been associated
        with a particular job at time of calling.

        @rtype:     string or None
        @return:    Return host uname or None.
        """
        raise NotImplementedError("uname not yet implemented")

    @property
    def total_ram(self):
        """
        Return the total ram of this host machine.
        Might be None if a host has not been associated
        with a particular job at time of calling.

        @rtype:     integer or None
        @return:    Return host RAM or None
        """
        raise NotImplementedError("total_ram not yet implemented")


class Task(WorkflowBase):
    """
    Class to expose information about a task associated
    with a particular job.  This is intended to be
    instantiated inside of a Job() object and not as a
    stand alone instance.

    Usage::

     t = Task()
     t.initialize('unique_wf_uuid', 3, 1)
     print t.start_time, t.duration

    etc
    """

    _indent = 3

    def __init__(self):
        WorkflowBase.__init__(self)

    def initialize(self, wf_id, job_id, task_id):
        """
        This method is the initialization method that accepts
        the unique wf_uuid, job_submit_seq and task_submit_seq
        that triggers the subclass specific queries and
        calculations that pulls task information from the back end.

        The wf_id is represented in the .bp logs as "wf.id".
        The job_id is represented as "job.id".
        The task_id is represented as "task.id".

        @type   wf_id: string
        @param  wf_id: the unique wf_uuid as defined by Pegasus
        @type   job_id: integer
        @param  job_id: the sequence number as defined
                by tailstatd
        @type   task_id: integer
        @param  task_id: the sequence number as defined
                by tailstatd
        """
        raise NotImplementedError("initialize not yet implemented")

    @property
    def task_submit_seq(self):
        """
        Return the task submit sequence number from the
        storage backend as an integer.

        @rtype:     int
        @return:    submit sequence number
        """
        raise NotImplementedError("task_submit_seq not yet implemented")

    @property
    def start_time(self):
        """
        Return start time of this task from the storage
        backend as a python datetime object (utc).

        @rtype:     python datetime obj (utc)
        @return:    Return task start time
        """
        raise NotImplementedError("start_time not yet implemented")

    @property
    def duration(self):
        """
        Return duration of this task from the storage
        backend as a float.

        @rtype:     float (from db)
        @return:    Return the duration of this task
        """
        raise NotImplementedError("duration not yet implemented")

    @property
    def exitcode(self):
        """
        Return the exitcode of this task from the storage
        backend as an integer.

        @rtype:     integer
        @return:    Return the task exitcode
        """
        raise NotImplementedError("exitcode not yet implemented")

    @property
    def transformation(self):
        """
        Return the transformation type of this task from
        the storage backend.

        @rtype:     string
        @return:    Return task transformation
        """
        raise NotImplementedError("transformation not yet implemented")

    @property
    def executable(self):
        """
        Return the executable invoked by this task from
        the storage backend.

        @rtype:     string
        @return:    Return the task executable
        """
        raise NotImplementedError("executable not yet implemented")

    @property
    def arguments(self):
        """
        Return the task args from the storage backend.

        @rtype:     string
        @return:    Return the task arguments
        """
        raise NotImplementedError("arguments not yet implemented")


# Base class for discovery/wf_uuid querying.


class Discovery:
    def __init__(self, connectionInfo=None):
        """
        Initialization method.  The manditory argument connectionInfo
        will be defined as appropriate in a subclass (string, dict, etc),
        and connect to the appropriate back end.
        """

    def fetch_all(self):
        """
        Void method that will return a list of workflow uuids
        from the back end.

        @rtype:     list
        @return:    Returns a list of wf_uuid strings.
        """
        raise NotImplementedError("fetch_all not yet implemented")

    def time_threshold(self, startTime=None, endTime=None):
        """
        Method to return a list of wf_uuid strings that
        were submitted within a certain timeframe.  The startTime
        arg should represent the time floor and endTime the time
        ceiling.  If both args are supplied return workflows
        that were exectued between the two.  If only one
        arg is supplied, use that one as the appropriate
        floor or ceiling.

        This is based on the workflowstate start event occurring
        after the startTime and/or before the endTime.

        @type       startTime: python datetime obj (utc)
        @param      startTime: The time "floor" to bound query.
        @type       endTime: python datetime obj (utc)
        @param      endTime: The time "ceiling" to bound query.
        @rtype:     list
        @return:    Returns a list of wf_uuid strings.
        """
        raise NotImplementedError("time_threshold not yet implemented")



import logging
import sys
from optparse import OptionParser

from Pegasus.tools import utils

log = logging.getLogger(__name__)


class Command:
    description = None
    epilog = None
    usage = "Usage: %prog [options] [args]"

    def __init__(self):
        self.parser = OptionParser(
            usage=self.usage, description=self.description, epilog=self.epilog
        )

    def parse(self, args):
        self.options, self.args = self.parser.parse_args(args)

    def run(self):
        pass

    def main(self, args=None):
        self.parse(args)
        self.run()


class LoggingCommand(Command):
    def __init__(self):
        Command.__init__(self)
        self.parser.add_option(
            "-v",
            "--verbose",
            action="count",
            default=0,
            dest="verbosity",
            help="Increase logging verbosity, repeatable",
        )

    def main(self, args=None):
        self.parse(args)

        verbosity = self.options.verbosity

        if verbosity == 0:
            log_level = logging.WARNING
        elif verbosity == 1:
            log_level = logging.INFO
        elif verbosity >= 2:
            log_level = logging.DEBUG

        utils.configureLogging(level=log_level)

        try:
            self.run()
        except Exception as e:
            # Only log stack grace if -v has been used
            if verbosity >= 1:
                log.exception(e)
            else:
                sys.stderr.write("%s\n" % e)
            exit(1)


class CompoundCommand(Command):
    "A Command with multiple sub-commands"
    usage = "%prog COMMAND [options] [args]"
    commands = []
    aliases = {}

    def __init__(self):
        Command.__init__(self)

        lines = ["\n\nCommands:"]
        for cmd, cmdclass in self.commands:
            lines.append("    {:<20} {}".format(cmd, cmdclass.description))

        self.parser.usage += "\n".join(lines)

    def main(self, args=None):
        if args is None:
            args = sys.argv[1:]

        command = None
        for arg in args:
            if arg[0] != "-":
                command = arg
                args.remove(command)
                break

        if command is None and ("-h" in args or "--help" in args):
            self.parser.print_help()
            exit(1)

        if command is None:
            self.parser.error("Specify COMMAND")
            exit(1)

        commandmap = dict(self.commands)

        if command in self.aliases:
            command = self.aliases[command]

        if command not in commandmap:
            sys.stderr.write("Invalid command: %s\n" % command)
            exit(1)

        cmdclass = commandmap[command]
        cmd = cmdclass()
        cmd.main(args)



#!/usr/bin/env python3

import logging
import os
import shutil
import subprocess
from json import dumps
from pathlib import Path

import click

from Pegasus.tools.utils import slurp_braindb

log = logging.getLogger("pegasus-run")


def check_dag(dag_file):
    """Check if DAG file exists."""
    dag_file = Path(dag_file)

    if not dag_file.is_file():
        raise FileNotFoundError("%s not found" % dag_file)
    elif not os.access(str(dag_file), os.R_OK):
        raise PermissionError("%s is not readable" % dag_file)

    log.debug("# found %s", dag_file)


def check_dag_sub_file(dag_sub_file):
    """Check if DAGman submit file exists."""
    dag_sub_file = Path(dag_sub_file)

    if not dag_sub_file.is_file():
        raise FileNotFoundError("%s not found" % dag_sub_file)
    elif not os.access(str(dag_sub_file), os.R_OK):
        raise PermissionError("%s is not readable" % dag_sub_file)

    log.debug("# dagman condor submit file is %s", dag_sub_file)


def get_condor_submit():
    """Locate `condor_submit` and check if it is executable."""
    condor_submit = shutil.which("condor_submit")

    if condor_submit is None:
        raise FileNotFoundError("condor_submit not found")
    elif not os.access(condor_submit, os.X_OK):
        raise PermissionError("%s is not executable" % condor_submit)

    log.debug("# found %s", condor_submit)
    return condor_submit


def get_grid_proxy_info(globus_location):
    """Locate `condor_submit` and check if it is executable."""
    grid_proxy_info = Path(globus_location) / "bin" / "grid-proxy-info"

    if not grid_proxy_info.exists():
        raise FileNotFoundError("%s not found" % grid_proxy_info)
    elif not os.access(str(grid_proxy_info), os.X_OK):
        raise PermissionError("%s is not executable" % grid_proxy_info)

    log.debug("# found %s", grid_proxy_info)
    return str(grid_proxy_info)


def salvage_log_file(condor_log):
    """Salvage jobs log files."""
    condor_log = Path(condor_log)

    if condor_log.exists():
        cl = str(condor_log)
        for i in range(1000):
            _cl = cl + ".%03d" % i
            if not Path(_cl).exists():
                log.debug("# log $result exists, rescuing from DAGMan.")
                try:
                    shutil.copyfile(cl, _cl)
                    click.secho("Rescued {} as {}".format(cl, _cl), err=True)
                except OSError as e:
                    raise ValueError(str(e))
                break


def grid_check():
    """Perform grid checks if enabled."""
    # sanity check: Is there a GLOBUS_LOCATION?
    globus_location = os.environ.get("GLOBUS_LOCATION", None)
    if globus_location is None:
        raise ValueError(
            "Your environment setup misses GLOBUS_LOCATION.\n"
            + "Please check carefully that you have sourced the correct setup files!",
        )

    log.debug("# GLOBUS_LOCATION=%s" % globus_location)

    # sanity check: Is GLOBUS_LOCCATION part of LD_LIBRARY_PATH?
    llp = os.environ.get("LD_LIBRARY_PATH", "")
    if not llp:
        llp = str(Path(globus_location) / "lib")
        os.putenv("LD_LIBRARY_PATH", llp)
        log.info("Setting LD_LIBRARY_PATH=%s", llp)

    # Find grid-proxy-info (should we use openssl instead? )
    gpi = get_grid_proxy_info(globus_location)

    # common user error
    # sanity check: Sufficient time left on grid proxy certificate
    cmd = (gpi, "-timeleft")

    rv = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    ec = rv.returncode

    if ec & 127:
        raise ValueError("%s died on signal %d" % (gpi, ec & 127))
    elif ec != 0:
        raise ValueError("%s exited with status %d" % (gpi, ec))

    time_left = int(rv.stdout.decode().strip())
    if time_left == -1:
        raise ValueError("Grid proxy not initialized, Please generate a new proxy")
    elif time_left == 0:
        raise ValueError("Grid proxy expired, please refresh")
    elif time_left < 7200:
        raise ValueError(
            "Too little time left (%d s) on grid proxy. Please refresh your proxy"
            % time_left
        )

    log.debug("# grid proxy has %d s left" % time_left)


def exec_dag(dag_sub_file, condor_log):
    """Execute `condor_submit` on dagman.condor.sub."""
    # PM-797 move away from using pegasus-submit-dag
    condor_submit = get_condor_submit()

    check_dag_sub_file(dag_sub_file)

    salvage_log_file(condor_log)
    click.secho("Submitting to condor %s" % dag_sub_file, err=True)
    cmd = (condor_submit, "-terse", dag_sub_file)

    rv = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    rv.check_returncode()
    return float(rv.stdout.decode().split("-")[0].strip())


def exec_script(script):
    """Locate `condor_submit` and check if it is executable."""
    script = Path(script)

    if not script.is_file():
        raise FileNotFoundError("%s not found" % script)
    elif not os.access(str(script), os.X_OK):
        raise PermissionError("%s is not executable" % script)

    log.debug("# found %s" % script)

    cmd = ("/bin/bash", str(script))
    rv = subprocess.run(cmd)
    rv.check_returncode()


@click.command()
@click.pass_context
@click.option(
    "--grid/--nogrid",
    default=False,
    show_default=True,
    help="Enable/Disable checking for grids.",
)
@click.option(
    "-j",
    "--json",
    default=False,
    show_default=True,
    is_flag=True,
    help="Output in JSON format.",
)
@click.option(
    "-v", "--verbose", default=0, count=True, help="Raises debug level by 1.",
)
@click.argument(
    "submit-dir",
    required=False,
    default=".",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, exists=True),
)
def pegasus_run(ctx, grid=False, json=False, verbose=0, submit_dir=None):
    """."""
    logging.basicConfig(level=logging.ERROR - (min(verbose, 3) * 10))

    os.umask(0o022)
    cwd = os.getcwd()
    config = slurp_braindb(submit_dir)
    submit_dir = str(Path(submit_dir).resolve())

    if not config:
        click.secho(
            click.style("Error: ", fg="red", bold=True)
            + "%s is not a valid submit-dir" % submit_dir
        )
        ctx.exit(1)

    try:
        os.chdir(submit_dir)
    except PermissionError:
        click.secho(
            click.style("Error: ", fg="red", bold=True)
            + "Cannot change to directory %s" % submit_dir
        )
        ctx.exit(1)

    if grid:
        try:
            grid_check()
        except (FileNotFoundError, PermissionError, ValueError) as e:
            click.secho(click.style("Error: ", fg="red", bold=True) + str(e))
            ctx.exit(1)

    if config["dag"]:
        try:
            # sanity check: Is the DAG file there?
            check_dag(config["dag"])

            # PM-870 we have already changed to the directory, don't prepend $run again
            dag_sub_file = config["dag"] + ".condor.sub"

            # PM-702: clean up .halt files from pegasus-halt
            halt_released = False
            if Path(config["dag"] + ".halt").exists():
                click.echo("Found a previously halted workflow. Releasing it now.")
                os.system("find . -name '*.dag.halt' -exec rm {} \\;")
                halt_released = True

            # After the switch from condor_submit_dag, we lost the check to see if
            # a workflow is already running. This replaces those checks.
            if Path("monitord.pid").exists():
                if halt_released:
                    ctx.exit(0)
                else:
                    click.secho(
                        click.style("Error: ", fg="red", bold=True)
                        + "It looks like the workflow is already running! If you are sure\n"
                        + "       that is not the case, please remove the monitord.pid file and try\n"
                        + "       again.",
                        err=True,
                    )
                    ctx.exit(1)

            # PM-797 do condor_submit on dagman.condor.sub file if it exists
            job_id = exec_dag(dag_sub_file, config["condor_log"])
            log.debug("# dagman is running")
            if json:
                click.echo(dumps(config))
            else:
                click.secho(
                    """Submitting job(s).
1 job(s) submitted to cluster %(job_id)d."""
                    % {"job_id": job_id,},
                    err=True,
                )
                click.secho(
                    """
Your workflow has been started and is running in the base directory:

%(submit_dir)s

*** To monitor the workflow you can run ***

pegasus-status -l %(submit_dir)s

*** To remove your workflow run ***

pegasus-remove %(submit_dir)s"""
                    % {"submit_dir": submit_dir}
                )
        except (FileNotFoundError, PermissionError, ValueError) as e:
            click.secho(click.style("Error: ", fg="red", bold=True) + str(e))
            ctx.exit(1)
        except subprocess.CalledProcessError as e:
            rc = e.returncode
            if rc != 0:
                click.secho(
                    click.style("Error: ", fg="red", bold=True)
                    + "Running %s failed with %d" % (e.cmd, rc)
                )
                ctx.exit(rc)

    elif config["type"] == "shell":
        try:
            if json:
                click.secho(
                    click.style("Error: ", fg="red", bold=True)
                    + "--json option is not supported for shell code"
                )
                ctx.exit(1)

            exec_script(config["script"])
            click.secho("✨ Success", fg="green")
        except (FileNotFoundError, PermissionError) as e:
            click.secho(click.style("Error: ", fg="red", bold=True) + str(e))
            ctx.exit(1)
        except subprocess.CalledProcessError as e:
            rc = e.returncode
            if rc != 0:
                click.secho(
                    click.style("Error: ", fg="red", bold=True)
                    + "Running %s failed with %d" % (config["script"], rc)
                )
                ctx.exit(rc)

    os.chdir(cwd)


if __name__ == "__main__":
    pegasus_run()



import calendar
import datetime
import re
from copy import copy
from optparse import Option, OptionValueError


def check_magicdate(option, opt, value):
    try:
        return magicdate(value)
    except Exception:
        raise OptionValueError("option {}: invalid date value: {!r}".format(opt, value))


class MagicDateOption(Option):
    TYPES = Option.TYPES + ("magicdate",)
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["magicdate"] = check_magicdate


res = [
    # x time ago
    (
        re.compile(
            r"""^
                ((?P<weeks>\d+) \s weeks?)?
                [^\d]*
                ((?P<days>\d+) \s days?)?
                [^\d]*
                ((?P<hours>\d+) \s hours?)?
                [^\d]*
                ((?P<minutes>\d+) \s minutes?)?
                [^\d]*
                ((?P<seconds>\d+) \s seconds?)?
                \s
                ago
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.datetime.today()
        - datetime.timedelta(
            days=int(m.group("days") or 0),
            seconds=int(m.group("seconds") or 0),
            minutes=int(m.group("minutes") or 0),
            hours=int(m.group("hours") or 0),
            weeks=int(m.group("weeks") or 0),
        ),
    ),
    # Today
    (
        re.compile(
            r"""^
                tod                             # Today
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today(),
    ),
    # Now (special case, returns datetime.datetime
    (
        re.compile(
            r"""^
                now                             # Now
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.datetime.now(),
    ),
    # Tomorrow
    (
        re.compile(
            r"""^
                tom                             # Tomorrow
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today() + datetime.timedelta(days=1),
    ),
    # Yesterday
    (
        re.compile(
            r"""^
                yes                             # Yesterday
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today() - datetime.timedelta(days=1),
    ),
    # 4th
    (
        re.compile(
            r"""^
                (?P<day>\d{1,2})                # 4
                (?:st|nd|rd|th)?                # optional suffix
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today().replace(day=int(m.group("day"))),
    ),
    # 4th Jan
    (
        re.compile(
            r"""^
                (?P<day>\d{1,2})                # 4
                (?:st|nd|rd|th)?                # optional suffix
                \s+                             # whitespace
                (?P<month>\w+)                  # Jan
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today().replace(
            day=int(m.group("day")), month=_parseMonth(m.group("month"))
        ),
    ),
    # 4th Jan 2003
    (
        re.compile(
            r"""^
                (?P<day>\d{1,2})                # 4
                (?:st|nd|rd|th)?                # optional suffix
                \s+                             # whitespace
                (?P<month>\w+)                  # Jan
                ,?                              # optional comma
                \s+                             # whitespace
                (?P<year>\d{4})                 # 2003
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=_parseMonth(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # Jan 4th
    (
        re.compile(
            r"""^
                (?P<month>\w+)                  # Jan
                \s+                             # whitespace
                (?P<day>\d{1,2})                # 4
                (?:st|nd|rd|th)?                # optional suffix
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date.today().replace(
            day=int(m.group("day")), month=_parseMonth(m.group("month"))
        ),
    ),
    # Jan 4th 2003
    (
        re.compile(
            r"""^
                (?P<month>\w+)                  # Jan
                \s+                             # whitespace
                (?P<day>\d{1,2})                # 4
                (?:st|nd|rd|th)?                # optional suffix
                ,?                              # optional comma
                \s+                             # whitespace
                (?P<year>\d{4})                 # 2003
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=_parseMonth(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # mm/dd/yyyy (American style, default in case of doubt)
    (
        re.compile(
            r"""^
                (?P<month>0?[1-9]|10|11|12)     # m or mm
                /                               #
                (?P<day>0?[1-9]|[12]\d|30|31)   # d or dd
                /                               #
                (?P<year>\d{4})                 # yyyy
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=int(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # dd/mm/yyyy (European style)
    (
        re.compile(
            r"""^
                (?P<day>0?[1-9]|[12]\d|30|31)   # d or dd
                /                               #
                (?P<month>0?[1-9]|10|11|12)     # m or mm
                /                               #
                (?P<year>\d{4})                 # yyyy
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=int(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # yyyy-mm-dd (ISO style)
    (
        re.compile(
            r"""^
                (?P<year>\d{4})                 # yyyy
                -                               #
                (?P<month>0?[1-9]|10|11|12)     # m or mm
                -                               #
                (?P<day>0?[1-9]|[12]\d|30|31)   # d or dd
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=int(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # yyyymmdd
    (
        re.compile(
            r"""^
                (?P<year>\d{4})                 # yyyy
                (?P<month>0?[1-9]|10|11|12)     # m or mm
                (?P<day>0?[1-9]|[12]\d|30|31)   # d or dd
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: datetime.date(
            year=int(m.group("year")),
            month=int(m.group("month")),
            day=int(m.group("day")),
        ),
    ),
    # next Tuesday
    (
        re.compile(
            r"""^
                next                            # next
                \s+                             # whitespace
                (?P<weekday>\w+)                # Tuesday
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: _nextWeekday(_parseWeekday(m.group("weekday"))),
    ),
    # last Tuesday
    (
        re.compile(
            r"""^
                (last                           # last
                \s+)?                           # whitespace
                (?P<weekday>\w+)                # Tuesday
                $                               # EOL
            """,
            (re.VERBOSE | re.IGNORECASE),
        ),
        lambda m: _lastWeekday(_parseWeekday(m.group("weekday"))),
    ),
]


def _parseMonth(input):
    months = "January February March April May June July August September October November December".split(
        " "
    )
    for i, month in enumerate(months):
        p = re.compile(input, re.IGNORECASE)
        if p.match(month):
            return i + 1
    else:
        raise Exception


def _parseWeekday(input):
    days = "Monday Tuesday Wednesday Thursday Friday Saturday Sunday".split(" ")
    for i, day in enumerate(days):
        p = re.compile(input, re.IGNORECASE)
        if p.match(day):
            return i
    else:
        raise Exception


def _nextWeekday(weekday):
    day = datetime.date.today() + datetime.timedelta(days=1)
    while calendar.weekday(*day.timetuple()[:3]) != weekday:
        day = day + datetime.timedelta(days=1)

    return day


def _lastWeekday(weekday):
    day = datetime.date.today() - datetime.timedelta(days=1)
    while calendar.weekday(*day.timetuple()[:3]) != weekday:
        day = day - datetime.timedelta(days=1)

    return day


def magicdate(input):
    for r, f in res:
        m = r.match(input.strip())
        if m:
            return f(m)



__author__ = "Monte Goode"

import logging
import time

from sqlalchemy import orm

from Pegasus.db import connection
from Pegasus.db.schema import *

log = logging.getLogger(__name__)

"""
Module to expunge a workflow and the associated data from
a stampede schema database in the case of running with the replay
option or a similar situation.

The wf_uuid that is passed into the constructor MUST be the
"top-level" workflow the user wants to delete.  Which is to
say if the wf_uuid is a the child of another workflow, then
only the data associated with that workflow will be deleted.
Any parent or sibling workflows will be left untouched.

Usage::

 from Pegasus.db import expunge

 connString = 'sqlite:///pegasusMontage.db'
 wf_uuid = '1249335e-7692-4751-8da2-efcbb5024429'
 expunge.delete_workflow(connString, wf_uuid)

All children/grand-children/etc information and associated
workflows will be removed.
"""


def delete_workflow(dburi, wf_uuid):
    "Expunge workflow from workflow database"

    log.info("Expunging %s from workflow database", wf_uuid)

    session = connection.connect(dburi, create=True)
    try:
        query = session.query(Workflow).filter(Workflow.wf_uuid == wf_uuid)
        try:
            wf = query.one()
        except orm.exc.NoResultFound as e:
            log.warn("No workflow found with wf_uuid %s - aborting expunge", wf_uuid)
            return

        # PM-1218 gather list of descendant workflows with wf_uuid
        query = session.query(Workflow).filter(Workflow.root_wf_id == wf.wf_id)
        try:
            desc_wfs = query.all()
            for desc_wf in desc_wfs:
                # delete the files from the rc_lfn explicitly as they are
                # not associated with workflow table
                __delete_workflow_files__(session, desc_wf.wf_uuid, desc_wf.wf_id)
        except orm.exc.NoResultFound as e:
            log.warn(
                "No workflow found with root wf_id %s - aborting expunge", wf.wf_id
            )
            return

        session.delete(wf)

        log.info("Flushing top-level workflow: %s", wf.wf_uuid)
        i = time.time()
        session.flush()
        session.commit()
        log.info("Flush took: %f seconds", time.time() - i)
    finally:
        session.close()


def __delete_workflow_files__(session, wf_uuid, wf_id):
    # Expunge all files associated with the workflow from the rc tables
    log.info(
        "Expunging rc files for workflow %s with database id %s from workflow database"
        % (wf_uuid, wf_id)
    )

    query = session.query(RCLFN).filter(
        RCLFN.lfn_id.in_(
            session.query(WorkflowFiles.lfn_id).filter(WorkflowFiles.wf_id == wf_id)
        )
    )
    count = query.delete(synchronize_session=False)
    log.info("Flushing deletes of rc_lfn from workflow: %s", wf_uuid)
    i = time.time()
    session.flush()
    session.commit()
    log.info("Flush took: %f seconds", time.time() - i)
    log.info(
        "Deleted  {} files from rc_file table for workflow {} ".format(count, wf_uuid)
    )


def delete_dashboard_workflow(dburi, wf_uuid):
    "Expunge workflow from dashboard database"

    log.info("Expunging %s from dashboard database", wf_uuid)

    session = connection.connect(dburi, create=True)
    try:
        query = session.query(MasterWorkflow).filter(MasterWorkflow.wf_uuid == wf_uuid)
        try:
            wf = query.one()
        except orm.exc.NoResultFound as e:
            log.warn("No workflow found with wf_uuid %s - aborting expunge", wf_uuid)
            return

        session.delete(wf)

        i = time.time()
        session.flush()
        session.commit()
        log.info("Flush took: %f seconds", time.time() - i)
    finally:
        session.close()



## Copyright (c) 2004, The Regents of the University of California, through
## Lawrence Berkeley National Laboratory (subject to receipt of any required
## approvals from the U.S. Dept. of Energy).  All rights reserved.

"""
NetLogger instrumentation API for Python

Write NetLogger log messages. Most users of this API will
use the Log class, which is a little like a 'Logger' object in the
Python logging API.

Utility functions include functions to get and set the Grid Job ID.
"""
__author__ = "Dan Gunter"
__created__ = "1 April 2004"
__rcsid__ = "$Id: nlapi.py 27037 2011-02-04 20:16:27Z dang $"

import calendar
import datetime
import os
import socket
import sys
import time
import urllib.parse

from Pegasus.netlogger.nldate import utcFormatISO
from Pegasus.netlogger.util import uuid1

#
## Exceptions
#


class ParseException(Exception):
    pass


class FormatException(Exception):
    pass


#
## Constants
#

# Environment variable to store GID
GID_ENV = "NETLOGGER_GUID"  # new
# OLD: GID_ENV = 'NL_GID'

# Environment variable to store destination
NLDEST_ENV = "NL_DEST"

# Environment variable for level file
CFG_ENV = "NL_CFG"

FIELD_SEP = " "
REC_SEP = "\n"
EOR = "\n"
KEYVAL_SEP = "="

# Port
DEFAULT_PORT = 14380

# Level
class Level:
    NOLOG = 0
    FATAL = 1
    ERROR = 2
    WARN = 3
    WARNING = 3
    INFO = 4
    DEBUG = 5
    DEBUG1 = 6
    DEBUG2 = 7
    DEBUG3 = 8
    TRACE = DEBUG1
    ALL = -1

    names = {
        NOLOG: "NOLOG",
        FATAL: "Fatal",
        ERROR: "Error",
        WARN: "Warn",
        INFO: "Info",
        DEBUG: "Debug",
        TRACE: "Trace",
        DEBUG2: "Debug2",
        DEBUG3: "Debug3",
    }

    @staticmethod
    def getName(level):
        return Level.names.get(level, "User")

    @staticmethod
    def getLevel(name):
        if name.isupper() and hasattr(Level, name):
            return getattr(Level, name)
        raise ValueError("no such level name: %s" % name)


DATE_FMT = "%04d-%02d-%02dT%02d:%02d:%02d"

# Required fields
TS_FIELD = "ts"
EVENT_FIELD = "event"
# Other conventions
LEVEL_FIELD = "level"
STATUS_FIELD = "status"
MESSAGE_FIELD = "msg"
HASH_FIELD = "nlhash"

#
## Utility functions
#


def quotestr(v):
    """Quote a string value to be output."""
    if not v:
        v = '""'
    elif " " in v or "\t" in v or '"' in v or "=" in v:
        v = '"%s"' % v.replace(r'"', r"\"")
    return v


def getGuid(create=True, env=GID_ENV):
    """Return a GUID.
    If 'create' is True (the default), and if none is found
    in the environment then create one.
    """
    gid = os.environ.get(env, None)
    if gid is None:
        if create:
            gid = uuid1()
    return gid


# Call this if you want to set a GID manually
def setGuid(id, env=GID_ENV):
    """Replace current guid in the environment with provided value.
    Return old value, or None if there was no old value.

    Note: may cause memory leak on FreeBSD and MacOS. See system docs.
    """
    old_gid = os.environ.get(env, None)
    os.environ[env] = id
    return old_gid


def clearGuid(env=GID_ENV):
    """Unset guid"""
    old_gid = os.environ.get(env, None)
    if old_gid:
        del os.environ[env]
    return old_gid


_g_hostip = None


def getHost():
    global _g_hostip
    if _g_hostip is not None:
        return _g_hostip
    try:
        ip = socket.gethostbyname(socket.getfqdn())
    except Exception:
        ip = "127.0.0.1"
    _g_hostip = ip
    return ip


def getProg():
    import sys

    return sys.argv[0]


def getDest():
    return os.environ.get(NLDEST_ENV, None)


class LevelConfig:
    """Set logging level from a configuration file.
    The format of the file is trivial: an integer log level.
    """

    DEFAULT = Level.INFO

    def __init__(self, filename):
        self._f = filename
        self._level = None

    def getLevel(self):
        if self._level is None:
            try:
                self._level = self.DEFAULT
                f = file(self._f)
                line = f.readline()
                i = int(line.strip())
                self._level = i
            except OSError:
                pass
            except ValueError:
                pass
        return self._level


if os.getenv(CFG_ENV) is not None:
    g_level_cfg = LevelConfig(os.getenv(CFG_ENV))
else:
    g_level_cfg = None


class Log:
    """NetLogger log class.

    Name=value pairs for the log are passed as keyword arguments.
    This is mostly good, but one drawback is that a period '.' in the
    name is confusing to python. As a work-around, use '__' to mean '.',
    e.g. if you want the result to be "foo.id=bar", then do::
        log.write(.., foo__id='bar')
    Similarly, a leading '__' will be stripped (e.g. to avoid stepping
    on keywords like 'class')

    If you instantiate this class without a 'logfile', it will act
    as a formatter, returning a string.

    To disable filtering of messages on level, add 'level=Level.ALL'
    """

    class OpenError(Exception):
        pass

    def __init__(
        self,
        logfile=None,
        flush=False,
        prefix=None,
        level=Level.INFO,
        newline=True,
        guid=True,
        pretty=False,
        float_time=False,
        meta={},
    ):
        """Constructor."""
        self._logfile = None
        self._float_time = float_time
        self._pretty = pretty
        self._newline = newline
        self._flush = [None, self.flush][flush]
        self.setPrefix(prefix)
        self._meta = {}
        if meta:
            self._meta[None] = meta
        if isinstance(logfile, str):
            try:
                self._logfile = urlfile(logfile)
            except (socket.gaierror, OSError) as E:
                raise self.OpenError(E)
        else:
            self._logfile = logfile
        if g_level_cfg is None:
            self._level = level
        else:
            self._level = g_level_cfg.getLevel()
        if guid is True:
            guid = getGuid(create=False)
            if guid:
                _m = self._meta.get(None, {})
                _m["guid"] = guid
                self._meta[None] = _m
        elif isinstance(guid, str):
            _m = self._meta.get(None, {})
            _m["guid"] = guid
            self._meta[None] = _m

    def setLevel(self, level):
        """Set highest level of messages that WILL be logged.
        Messages below this level (that is, less severe,
        higher numbers) will be dropped.

        For example::
          log.setLevel(Level.WARN)
          log.error('argh',{}) # logged
          log.info('whatever',{}) # dropped!
        """
        self._level = level

    def setPrefix(self, prefix):
        if prefix is None:
            self._pfx = ""
        elif prefix.endswith("."):
            self._pfx = prefix
        else:
            self._pfx = prefix + "."

    def debugging(self):
        """Return whether the level >= debug."""
        return self._level >= Level.DEBUG

    def flush(self):
        """Flush output object."""
        if self._logfile:
            self._logfile.flush()

    def write(self, event="event", ts=None, level=Level.INFO, **kw):
        """Write a NetLogger string.
        If there is a logfile, returns None
        Otherwise, returns a string that would have been written.
        """
        if self._level != Level.ALL and level > self._level:
            if self._logfile:
                return None
            else:
                return ""
        if not ts:
            ts = time.time()
        buf = self.format(self._pfx + event, ts, level, kw)
        if self._logfile is None:
            return buf
        self._logfile.write(buf)
        if self._flush:
            self.flush()

    __call__ = write

    def error(self, event="", **kwargs):
        return self.write(event, level=Level.ERROR, **kwargs)

    def warn(self, event="", **kwargs):
        return self.write(event, level=Level.WARN, **kwargs)

    def info(self, event="", **kwargs):
        return self.write(event, level=Level.INFO, **kwargs)

    def debug(self, event="", **kwargs):
        return self.write(event, level=Level.DEBUG, **kwargs)

    def _append(self, fields, kw):
        for k, v in kw.items():
            if k.startswith("__"):
                k = k[2:]
            k = k.replace("__", ".")
            if isinstance(v, str):
                v = quotestr(v)
                fields.append("{}={}".format(k, v))
            elif isinstance(v, float):
                fields.append("{}={:f}".format(k, v))
            elif isinstance(v, int):
                fields.append("%s=%d" % (k, v))
            else:
                s = str(v)
                if " " in s or "\t" in s:
                    s = '"%s"' % s
                fields.append("{}={}".format(k, s))

    def format(self, event, ts, level, kw):
        if not self._pretty:
            # Regular BP formatting
            if isinstance(ts, str):
                fields = ["ts=" + ts, "event=" + event]
            elif isinstance(ts, datetime.datetime):
                if self._float_time:
                    tsfloat = calendar.timegm(ts.utctimetuple()) + ts.microsecond / 1e6
                    fields = ["ts=%.6f" % tsfloat, "event=" + event]
                else:
                    tsstr = "%s.%06dZ" % (
                        DATE_FMT % ts.utctimetuple()[0:6],
                        ts.microsecond,
                    )
                    fields = ["ts=" + tsstr, "event=" + event]
            elif self._float_time:
                fields = ["ts=%.6f" % ts, "event=" + event]
            else:
                fields = ["ts=" + utcFormatISO(ts), "event=" + event]
            if level is not None:
                if isinstance(level, int):
                    fields.append("level=" + Level.getName(level))
                else:
                    fields.append("level=%s" % level)
            if kw:
                self._append(fields, kw)
            if event in self._meta:
                self._append(fields, self._meta[event])
            if None in self._meta:
                self._append(fields, self._meta[None])
            buf = FIELD_SEP.join(fields)
        else:
            # "Pretty" BP formatting
            if not isinstance(ts, str):
                ts = utcFormatISO(ts)
            if isinstance(level, int):
                level = Level.getName(level).upper()
            # print traceback later
            if "traceback" in kw:
                tbstr = kw["traceback"]
                del kw["traceback"]
            else:
                tbstr = None
            if "msg" in kw:
                msg = kw["msg"]
                del kw["msg"]
            else:
                msg = None
            remainder = ",".join(
                "{}={}".format(key, value) for key, value in kw.items()
            )
            if msg:
                buf = "{} {:<6} {} | {}. {}".format(ts, level, event, msg, remainder)
            else:
                buf = "{} {:<6} {} | {}".format(ts, level, event, remainder)
            # add traceback
            if tbstr:
                buf += "\n" + tbstr
        if self._newline:
            return buf + REC_SEP
        else:
            return buf

    def setMeta(self, event=None, **kw):
        self._meta[event] = kw

    def close(self):
        self.flush()

    def __del__(self):
        if not hasattr(self, "closed"):
            self.close()
        self.closed = True

    def __str__(self):
        if self._logfile:
            return str(self._logfile)
        else:
            return repr(self)


# set up urlparse to recognize x-netlog schemes
for scheme in "x-netlog", "x-netlog-udp":
    urllib.parse.uses_netloc.append(scheme)
    try:
        urllib.parse.uses_query.append(scheme)
    except AttributeError:
        pass


def urlfile(url):
    """urlfile(url:str) -> file

    Open a NetLogger URL and return a write-only file object.
    """
    # print "url='%s'" % url
    # Split URL
    scheme, netloc, path, params, query, frag = urllib.parse.urlparse(url)
    # Put query parts into a dictionary for easy access later
    query_data = {}
    if query:
        query_parts = query.split("&")
        for flag in query_parts:
            name, value = flag.split("=")
            query_data[name] = value
    # Create file object
    if scheme == "file" or scheme == "" or scheme is None:
        # File
        if path == "-":
            fileobj = sys.stdout
        elif path == "&":
            fileobj = sys.stderr
        else:
            if "append" in query_data:
                is_append = boolparse(query_data["append"])
                open_flag = "aw"[is_append]
            else:
                open_flag = "a"
            fileobj = file(path, open_flag)
    elif scheme.startswith("x-netlog"):
        # TCP or UDP socket
        if netloc.find(":") == -1:
            addr = (netloc, DEFAULT_PORT)
        else:
            host, port_str = netloc.split(":")
            addr = (host, int(port_str))
        if scheme == "x-netlog":
            # TCP Socket
            sock = socket.socket()
        elif scheme == "x-netlog-udp":
            # UDP Socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        else:
            raise ValueError(
                "Unknown URL scheme '%s', "
                "must be empty, 'file' or 'x-netlog[-udp]'" % scheme
            )
        # print "connect to address %s" % addr
        sock.connect(addr)
        fileobj = sock.makefile("w")
    else:
        raise ValueError(
            "Unknown URL scheme '%s', "
            "must be empty, 'file' or 'x-netlog[-udp]'" % scheme
        )
    return fileobj


def urltype(url):
    """urltype(url:str) -> 'file' | 'tcp' | None

    Return a canonical string representing the type of URL,
    or None if the type is unknown
    """
    scheme = urllib.parse.urlparse(url)[0]
    if scheme == "file" or scheme == "" or scheme is None:
        return "file"
    elif scheme == "x-netlog":
        return "tcp"
    else:
        return None


# Get host

_g_hostip = None


def get_host():
    global _g_hostip
    if _g_hostip is not None:
        return _g_hostip
    try:
        ip = socket.gethostbyname(socket.getfqdn())
    except Exception:
        ip = "127.0.0.1"
    _g_hostip = ip
    return ip



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import datetime
import logging
import os
import shutil
import subprocess
import sys
import time
import warnings

from sqlalchemy import func
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm.exc import NoResultFound

from Pegasus.db import connection
from Pegasus.db.schema import *
from Pegasus.db.schema import check_table_exists, get_missing_tables, metadata

__author__ = "Rafael Ferreira da Silva"

log = logging.getLogger(__name__)

# -------------------------------------------------------------------
# DB Admin configuration
# -------------------------------------------------------------------
CURRENT_DB_VERSION = 14
DB_MIN_VERSION = 8

COMPATIBILITY = {
    "4.3.0": 1,
    "4.3.1": 1,
    "4.3.2": 1,
    "4.4.0": 2,
    "4.4.1": 2,
    "4.4.2": 2,
    "4.5.0": 4,
    "4.5.1": 4,
    "4.5.2": 4,
    "4.5.3": 4,
    "4.5.4": 5,
    "4.6.0": 6,
    "4.6.1": 6,
    "4.6.2": 6,
    "4.7.0": 8,
    "4.7.3": 8,
    "4.8.0": 8,
    "4.8.1": 8,
    "4.8.2": 8,
    "4.8.3": 8,
    "4.9.0": 11,
    "4.9.1": 11,
    "4.9.2": 11,
    "4.9.3": 11,
    "5.0.0": 14,
    "5.0.1": 14,
}


# -------------------------------------------------------------------


class DBAdminError(Exception):
    def __init__(self, message, db=None, db_version=None, given_version=None):
        """
        :param message: Exception message
        :param db: DB session object
        :param db_version: Current DB version (integer)
        :param given_version: Provided pegasus version
        """
        super().__init__(message)

        self.db = db
        self.db_version = db_version
        self.given_version = given_version

        if db:
            self.dburi = db.get_bind().url
            if not db_version:
                try:
                    self.db_version = get_version(db)
                except (NoResultFound, ProgrammingError):
                    pass
            self.db_compatible_version = get_compatible_version(self.db_version)

            if given_version:
                self.pegasus_version = COMPATIBILITY[given_version]
                self.pegasus_compatible_version = given_version
            else:
                self.pegasus_version = CURRENT_DB_VERSION
                self.pegasus_compatible_version = get_compatible_version(
                    CURRENT_DB_VERSION
                )


def get_compatible_version(version):
    """
    Get a compatible Pegasus version for the database version.
    :param version: version of the database
    :return: the equivalent Pegasus version
    """
    if version == CURRENT_DB_VERSION:
        pegasus_version = None

        # PM-1596 - Python codes now get PEGASUS_HOME set correctly
        if "PEGASUS_HOME" in os.environ:
            f = os.path.join(os.environ.get("PEGASUS_HOME"), "bin/pegasus-version")
            if os.path.isfile(f):
                pegasus_version = f

        if not pegasus_version:
            f = os.path.join(os.path.dirname(sys.argv[0]), "pegasus-version")
            if os.path.isfile(f):
                pegasus_version = f

        if pegasus_version:
            child = subprocess.Popen(
                pegasus_version,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                cwd=os.getcwd(),
            )
            out, err = child.communicate()
            if child.returncode != 0:
                raise DBAdminError(err.decode("utf8").strip())
            return out.decode("utf8").strip()

    print_version = None
    previous_version = None

    if version is not None and int(version) >= CURRENT_DB_VERSION:
        pv = -1
        for ver in COMPATIBILITY:
            if COMPATIBILITY[ver] > pv and (
                previous_version is None or ver > previous_version
            ):
                pv = COMPATIBILITY[ver]
                print_version = ver
                previous_version = ver

    for ver in COMPATIBILITY:
        if COMPATIBILITY[ver] == version and (
            previous_version is None or ver > previous_version
        ):
            print_version = ver
            previous_version = ver
    return print_version


def get_class(version, db):
    """
    Get a database version class for update/downgrade.
    :param version: version of the database
    :param db: DB session object
    :return: a database version class
    """
    module = "Pegasus.db.admin.versions.v{}".format(version)
    mod = __import__(module, fromlist=["Version"])
    klass = getattr(mod, "Version")
    return klass(db)


# -------------------------------------------------------------------
def db_create(
    dburi,
    engine,
    db,
    pegasus_version=None,
    force=False,
    verbose=True,
    print_version=True,
):
    """
    Create/Update the Pegasus database from the schema.
    :param dburi: URL to the db
    :param engine: DB engine object
    :param db: DB session object
    :param pegasus_version: version of the Pegasus software (e.g., 5.0.1)
    :param force: whether operations should be performed despite conflicts
    :param verbose: whether messages should be printed in the prompt
    :param print_version: whether to print the database version
    """
    table_names = engine.table_names()
    db_version = DBVersion.__table__
    db_version.create(engine, checkfirst=True)

    v = 0
    if len(table_names) == 0:
        engine.execute(
            db_version.insert(),
            version=CURRENT_DB_VERSION,
            version_number=int(CURRENT_DB_VERSION),
            version_timestamp=datetime.datetime.now().strftime("%s"),
        )
        if verbose:
            print("Pegasus database was successfully created.")
    else:
        v = _discover_version(
            db, pegasus_version=pegasus_version, force=force, verbose=False
        )

    try:
        metadata.create_all(engine, checkfirst=True)
    except OperationalError as e:
        raise DBAdminError(e, db=db, given_version=pegasus_version)

    if verbose and v > 0:
        print("Database has been updated: {}".format(dburi))

    print_db_version(print_version, v if v > 0 else CURRENT_DB_VERSION, db, parse=True)


def db_verify(db, check=False, pegasus_version=None, force=False, print_version=True):
    """
    Verify whether the database is compatible to the specified Pegasus version.
    :param db: DB session object
    :param check: whether to check database compatibility
    :param pegasus_version: version of the Pegasus software (e.g., 5.0.1)
    :param force: whether operations should be performed despite conflicts
    :param print_version: whether to print the database version
    """
    _has_version_table(db, pegasus_version)
    db_version = get_version(db)

    if check:
        _verify_tables(db, db_version)
        version = parse_pegasus_version(pegasus_version)

        if (
            db_version
            and db_version <= CURRENT_DB_VERSION
            and not version == db_version
        ):
            raise DBAdminError(
                "Database is NOT compatible with version: {} ({})".format(
                    get_compatible_version(version), db.get_bind().url
                ),
                db=db,
                db_version=db_version,
                given_version=pegasus_version,
            )

    print_db_version(print_version, db_version, db, parse=True)


def db_downgrade(db, pegasus_version=None, force=False, verbose=True):
    """
    Downgrade the database.
    :param db: DB session object
    :param pegasus_version: version of the Pegasus software (e.g., 4.6.0)
    :param force: whether operations should be performed despite conflicts
    :param verbose: whether messages should be printed in the prompt
    """
    _has_version_table(db, pegasus_version)
    try:
        current_version = get_version(db, sanity_check=False)
    except NoResultFound:
        raise DBAdminError(
            "Unable to determine database version.",
            db=db,
            given_version=pegasus_version,
        )

    if pegasus_version:
        version = parse_pegasus_version(pegasus_version)
    else:
        previous_version = ""
        for ver in COMPATIBILITY:
            if COMPATIBILITY[ver] < current_version and ver > previous_version:
                version = COMPATIBILITY[ver]
                previous_version = ver

    if current_version == version:
        log.info("Your database is already downgraded.")
        return
    elif current_version < version:
        raise DBAdminError(
            "Cannot downgrade to a higher version.",
            db=db,
            db_version=current_version,
            given_version=pegasus_version,
        )
    elif version < DB_MIN_VERSION:
        raise DBAdminError(
            "Cannot downgrade database to a version below Pegasus %s."
            % get_compatible_version(DB_MIN_VERSION),
            db=db,
            db_version=current_version,
            given_version=pegasus_version,
        )

    # backup the database before making changes
    _backup_db(db)

    for i in range(int(current_version), int(version) - 1, -1):

        if i == int(current_version):
            max_range = _get_minor_version(current_version)
        else:
            max_range = _get_max_minor_version(i)

        for j in range(max_range, 0, -1):
            k = get_class("{}-{}".format(i, j), db)
            k.downgrade(force)
            actual_version = float("{}.{}".format(i, j - 1))
            _update_version(db, actual_version)

        if i > version:
            k = get_class(i, db)
            k.downgrade(force)
            actual_version = float("{}.{}".format(i - 1, _get_max_minor_version(i - 1)))
            _update_version(db, actual_version)

        if actual_version == version:
            break

    if verbose:
        print("Your database was successfully downgraded.")


def parse_pegasus_version(pegasus_version=None):
    """
    Get database version associated to the Pegasus version.
    :param pegasus_version: version of the Pegasus software (e.g., 4.6.0)
    :return: database version
    """
    if pegasus_version == 0 or pegasus_version:
        if pegasus_version in COMPATIBILITY:
            return COMPATIBILITY[pegasus_version]
        raise DBAdminError(
            "Version does not exist: {}.".format(pegasus_version),
            given_version=pegasus_version,
        )
    return CURRENT_DB_VERSION


def all_workflows_db(
    db, update=True, pegasus_version=None, schema_check=True, force=False
):
    """
    Update/Downgrade all completed workflow databases listed in master_workflow table.
    :param db: DB session object
    :param pegasus_version: version of the Pegasus software (e.g., 5.0.1)
    :param schema_check: whether a sanity check of the schema should be performed
    :param force: whether operations should be performed despite conflicts
    """
    # log files
    file_prefix = "%s-dbadmin" % time.strftime("%Y%m%dT%H%M%S")
    f_out = open("%s.out" % file_prefix, "w")
    f_err = open("%s.err" % file_prefix, "w")

    data = (
        db.query(
            MasterWorkflow.db_url,
            MasterWorkflowstate.state,
            func.max(MasterWorkflowstate.timestamp),
        )
        .join(MasterWorkflowstate)
        .group_by(MasterWorkflow.wf_id)
        .all()
    )

    db_urls = []
    for d in data:
        if d[1] == "WORKFLOW_TERMINATED":
            db_urls.append(d[0])
            f_err.write("[ACTIVE] %s\n" % d[0])

    counts = {
        "total": len(data),
        "running": len(data) - len(db_urls),
        "success": 0,
        "failed": 0,
        "unable_to_connect": 0,
    }
    if update:
        msg = ["updating", "Updated"]
    else:
        msg = ["downgrading", "Downgraded"]

    print("")
    print("Verifying and %s workflow databases:" % msg[0])
    i = counts["running"]
    for dburi in db_urls:
        log.debug("{} '{}'...".format(msg[0], dburi))
        i += 1
        sys.stdout.write("\r%d/%d" % (i, counts["total"]))
        sys.stdout.flush()
        try:
            if update:
                con = connection.connect(
                    dburi,
                    pegasus_version=pegasus_version,
                    schema_check=schema_check,
                    create=True,
                    force=force,
                    verbose=False,
                )
            else:
                con = connection.connect(
                    dburi, schema_check=schema_check, create=False, verbose=False
                )
                metadata.clear()
                warnings.simplefilter("ignore")
                metadata.reflect(bind=con.get_bind())
                db_downgrade(
                    con, pegasus_version=pegasus_version, force=force, verbose=False
                )
            con.close()
            f_out.write("[SUCCESS] %s\n" % dburi)
            counts["success"] += 1
        except connection.ConnectionError as e:
            if "unable to open database file" in str(e):
                f_err.write("[UNABLE TO CONNECT] %s\n" % dburi)
                counts["unable_to_connect"] += 1
                log.debug(e)
            else:
                f_err.write("[ERROR] %s\n" % dburi)
                counts["failed"] += 1
                log.debug(e)
        except Exception as e:
            f_err.write("[ERROR] %s\n" % dburi)
            counts["failed"] += 1
            log.debug(e)

    f_out.close()
    f_err.close()

    print("\n\nSummary:")
    print("  Verified/{}: {}/{}".format(msg[1], counts["success"], counts["total"]))
    print("  Failed: {}/{}".format(counts["failed"], counts["total"]))
    print(
        "  Unable to connect: {}/{}".format(
            counts["unable_to_connect"], counts["total"]
        )
    )
    print(
        "  Unable to update (active workflows): %s/%s"
        % (counts["running"], counts["total"])
    )
    print("\nLog files:")
    print("  %s.out (Succeeded operations)" % file_prefix)
    print("  %s.err (Failed operations)" % file_prefix)


def print_db_version(print_version, db_version, db, parse=True):
    """
    Print the database version
    :param print_version: whether to print the database version
    :param db_version: Current DB version (integer)
    :param db: DB session object
    :param parse: whether database version should be presented as a version of the Pegasus software
    """
    if print_version:
        current_version = _db_current_version(db_version, db, parse=True)
        print("Database version: '{}' ({})".format(current_version, db.get_bind().url))


def get_version(db, sanity_check=True):
    """
    Get the DB version.
    :param db: DB session object
    :param sanity_check:
    :return: the DB current version (integer) or -1 if not found
    """
    try:
        current_version = (
            db.query(DBVersion.version).order_by(DBVersion.id.desc()).first()
        )
        if not current_version:
            log.debug("No version record found on dbversion table.")
            return -1

        if sanity_check:
            _version_sanity_check(db, current_version[0])

        return float(current_version[0])

    except OperationalError:
        return -1


################################################################################


def _db_current_version(current_version, db, parse=False):
    """
    Get the current version of the database.
    :param current_version:
    :param db: DB session object
    :param parse: whether database version should be presented as a version of the Pegasus software
    :return: current version of the database
    """
    if parse:
        current_version = get_compatible_version(current_version)
        if not current_version:
            raise DBAdminError(
                "Your database is not compatible with any Pegasus version.\nRun 'pegasus-db-admin "
                "update {}' to update it to the latest version.".format(
                    db.get_bind().url
                )
            )

    return current_version


def _discover_version(db, pegasus_version=None, force=False, verbose=True):
    """
    Discover Pegasus database version and update it if not the latest.
    :param db: DB session object
    :param pegasus_version: version of the Pegasus software (e.g., 5.0.1)
    :param force: whether operations should be performed despite conflicts
    :param verbose: whether messages should be printed in the prompt
    :return: database version
    """
    version = parse_pegasus_version(pegasus_version)

    current_version = 0
    if not force:
        current_version = get_version(db)

    if current_version == version:
        try:
            _verify_tables(db, current_version)
            log.debug("Database is already updated: {}".format(db.get_bind().url))
            return 0
        except DBAdminError:
            current_version = 0

    _version_sanity_check(db, current_version)

    if current_version > version:
        raise DBAdminError(
            "Unable to run update. Current database version is newer than specified version '{}'.".format(
                pegasus_version
            ),
            db=db,
            db_version=current_version,
            given_version=pegasus_version,
        )

    # update database
    _backup_db(db)
    v = 0.0
    for i in range(int(current_version), int(version) + 1):
        if not i == int(current_version):
            k = get_class(i, db)
            k.update(force=force)
        v = float(i)

        # verify minor versions
        max_range = 999
        if i == int(version):
            max_range = _get_minor_version(version)

        for j in range(1, max_range + 1):
            try:
                k = get_class("{}-{}".format(i, j), db)
                k.update(force=force)
                v = float("{}.{}".format(i, j))
            except ImportError:
                break

    if v > current_version:
        _update_version(db, v)
        if verbose:
            print("Database has been updated: {}".format(db.get_bind().url))
    else:
        v = 0
    return v


def _update_version(db, version):
    v = DBVersion()
    v.version = version
    v.version_number = int(version)
    v.version_timestamp = datetime.datetime.now().strftime("%s")
    if db:
        db.add(v)
        db.commit()


def _backup_db(db):
    """
    Create a copy of the database (SQLite), or create a dump of the database into a .sql file (MySQL).
    :param db: DB session object
    """
    url = db.get_bind().url
    # Backup SQLite database
    if url.drivername == "sqlite" and str(url).lower() != "sqlite://":
        dest_file = "{}-{}".format(url.database, time.strftime("%Y%m%d-%H%M%S"))
        shutil.copy(url.database, dest_file)
        log.debug("Created backup database file at: {}".format(dest_file))

    elif url.drivername == "mysql" or url.drivername == "postgresql":
        # Backup MySQL database
        if url.drivername == "mysql":
            log.info("Backing up MySQL database. This operation may take a while.")
            # mysqldump command preparation
            command = (
                "mysqldump"
                if not url.password
                else "export MYSQL_PWD={}; mysqldump".format(url.password)
            )
            if url.username:
                command += " -u {}".format(url.username)

        # Backup PostgreSQL database
        elif url.drivername == "postgresql":
            log.info("Backing up PostgreSQL database. This operation may take a while.")
            # pg_dump command preparation
            command = (
                "pg_dump"
                if not url.password
                else "export PGPASSWORD={}; pg_dump".format(url.password)
            )
            if url.username:
                command += " -U {}".format(url.username)

        if url.host:
            command += " -h {}".format(url.host)

        dest_file = "{}-{}.sql".format(url.database, time.strftime("%Y%m%d-%H%M%S"))
        command += " {} > {}".format(url.database, dest_file)
        child = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        out, err = child.communicate()
        if child.returncode != 0:
            if "Error 2013" in err:
                err += "\nPlease, refer to the pegasus-db-admin troubleshooting for possible ways to fix this error."
            raise DBAdminError(err, db=db)
        log.debug("Created backup database file at: {}".format(dest_file))


def _verify_tables(db, db_version=None):
    try:
        missing_tables = get_missing_tables(db)
    except Exception as e:
        raise DBAdminError(e, db_version=db_version, db=db)

    if len(missing_tables) > 0:
        raise DBAdminError(
            "Missing database tables or tables are not updated:\n    %s\n"
            "Run 'pegasus-db-admin update %s' to create/update your database."
            % (" \n    ".join(missing_tables), db.get_bind().url),
            db=db,
            db_version=db_version,
        )


def _get_minor_version(version):
    return int(str(float(version) - int(version))[2:])


def _get_max_minor_version(version):
    max_version = 0
    for ver in COMPATIBILITY:
        minor_version = _get_minor_version(COMPATIBILITY[ver])
        if int(COMPATIBILITY[ver]) == version and minor_version > max_version:
            max_version = minor_version
    return max_version


def _version_sanity_check(db, version):
    """
    Verify whether db version is higher than current version.
    :param db: db connection
    :param version: version to be verified
    """
    if float(version) > CURRENT_DB_VERSION:
        raise DBAdminError(
            "You database was created with a newer Pegasus version. "
            "It will not work properly with the current version."
            "\nPlease, run 'pegasus-db-admin downgrade' with the latest Pegasus to downgrade your "
            "database.",
            db=db,
            db_version=version,
        )


def _has_version_table(db, pegasus_version=None):
    """
    Verify whether version table exists.
    :param db: db connection
    :param pegasus_version: version of the Pegasus software (e.g., 5.0.1)
    """
    if not check_table_exists(db, DBVersion):
        raise DBAdminError(
            "Unable to determine database version.",
            db=db,
            db_version=CURRENT_DB_VERSION,
            given_version=pegasus_version,
        )



import logging

from flask import make_response
from sqlalchemy.orm.exc import NoResultFound

from Pegasus.service._query import InvalidQueryError
from Pegasus.service._serialize import jsonify
from Pegasus.service._sort import InvalidSortError
from Pegasus.service.base import ErrorResponse, InvalidJSONError
from Pegasus.service.monitoring import monitoring

log = logging.getLogger(__name__)

JSON_HEADER = {"Content-Type": "application/json"}
"""
Error

{
    "code"     : <string:code>,
    "message"  : <string:message>,
    "errors"   : [
        {
            "field"  : <string:field>,
            "errors" : [
                <string:errors>,
                ..
            ]
        },
        ..
    ]
}
"""


@monitoring.errorhandler(NoResultFound)
def no_result_found(error):
    e = ErrorResponse("NOT_FOUND", str(error))
    response_json = jsonify(e)

    return make_response(response_json, 404, JSON_HEADER)


@monitoring.errorhandler(InvalidQueryError)
def invalid_query_error(error):
    e = ErrorResponse("INVALID_QUERY", str(error))
    response_json = jsonify(e)

    return make_response(response_json, 400, JSON_HEADER)


@monitoring.errorhandler(InvalidSortError)
def invalid_order_error(error):
    e = ErrorResponse("INVALID_ORDER", str(error))
    response_json = jsonify(e)

    return make_response(response_json, 400, JSON_HEADER)


@monitoring.errorhandler(InvalidJSONError)
def invalid_json_error(error):
    e = ErrorResponse("INVALID_JSON", str(error))
    response_json = jsonify(e)

    return make_response(response_json, 400, JSON_HEADER)


@monitoring.errorhandler(Exception)
def catch_all(error):
    log.exception(error)

    app_code, http_code = error.codes if hasattr(error, "codes") else ("UNKNOWN", 500)

    e = ErrorResponse(app_code, str(error))
    response_json = jsonify(e)

    return make_response(response_json, http_code, JSON_HEADER)



"""."""

import glob
import logging
import os
import shutil
import tarfile
from os.path import expanduser

from Pegasus.command import CompoundCommand, LoggingCommand
from Pegasus.db import connection
from Pegasus.db.schema import (
    EnsembleWorkflow,
    MasterWorkflow,
    MasterWorkflowstate,
    Workflow,
    Workflowstate,
)
from Pegasus.tools import utils

log = logging.getLogger(__name__)


class SubmitDirException(Exception):
    pass


class MasterDatabase:
    def __init__(self, session):
        self.session = session

    def get_master_workflow(self, wf_uuid, submit_dir=None):
        q = self.session.query(MasterWorkflow)
        q = q.filter(MasterWorkflow.wf_uuid == wf_uuid)

        if submit_dir:
            q = q.filter(MasterWorkflow.submit_dir == submit_dir)

        wf = q.first()
        return wf

    def get_master_workflow_for_submitdir(self, submitdir):
        q = self.session.query(MasterWorkflow)
        q = q.filter(MasterWorkflow.submit_dir == submitdir)
        return q.all()

    def get_ensemble_workflow(self, wf_uuid):
        q = self.session.query(EnsembleWorkflow)
        q = q.filter(EnsembleWorkflow.wf_uuid == wf_uuid)
        return q.first()

    def delete_master_workflow(self, wf_uuid, submit_dir=None):
        w = self.get_master_workflow(wf_uuid, submit_dir=submit_dir)
        if w is None:
            return

        # Delete any ensemble workflows
        q = self.session.query(EnsembleWorkflow)
        q = q.filter(EnsembleWorkflow.wf_uuid == wf_uuid)
        q.delete()

        # Delete the workflow
        q = self.session.query(MasterWorkflow)
        q = q.filter(MasterWorkflow.wf_id == w.wf_id)
        q.delete()


class WorkflowDatabase:
    def __init__(self, session):
        self.session = session

    def delete_workflow(self, wf_uuid):
        q = self.session.query(Workflow)
        q = q.filter(Workflow.wf_uuid == wf_uuid)
        w = q.first()

        # If not found, do nothing
        if w is None:
            log.warning("Workflow not found in workflow DB: %s" % wf_uuid)
            return

        # Delete it
        self.session.delete(w)

    def get_workflow(self, wf_uuid):
        q = self.session.query(Workflow)
        q = q.filter(Workflow.wf_uuid == wf_uuid)
        return q.first()

    def get_workflow_states(self, wf_id):
        q = self.session.query(Workflowstate)
        q = q.filter(Workflowstate.wf_id == wf_id)
        return q.all()

    def update_submit_dirs(self, root_wf_id, src, dest):
        q = self.session.query(Workflow)
        q = q.filter(Workflow.root_wf_id == root_wf_id)
        for wf in q.all():
            log.info("Old submit dir: %s" % wf.submit_dir)
            wf.submit_dir = wf.submit_dir.replace(src, dest)
            log.info("New submit dir: %s" % wf.submit_dir)


class SubmitDir:
    def __init__(self, submitdir, raise_err=True):
        self.submitdir = os.path.abspath(submitdir)
        self.submitdir_exists = True

        if not os.path.isdir(submitdir):
            self.submitdir_exists = False

            if raise_err is False:
                return

            raise SubmitDirException("Invalid submit dir: %s" % submitdir)

        self.braindump_file = os.path.join(self.submitdir, "braindump.yml")
        if not os.path.isfile(self.braindump_file):
            self.braindump_file = os.path.join(self.submitdir, "braindump.txt")

        # Read the braindump file
        self.braindump = utils.slurp_braindb(os.path.join(self.submitdir))

        # Read some attributes from braindump file
        self.wf_uuid = self.braindump["wf_uuid"]
        self.root_wf_uuid = self.braindump["root_wf_uuid"]
        self.user = self.braindump["user"]

        self.archname = os.path.join(self.submitdir, "archive.tar.gz")

    def is_subworkflow(self):
        "Check to see if this workflow is a subworkflow"
        return self.wf_uuid != self.root_wf_uuid

    def is_archived(self):
        "A submit dir is archived if the archive file exists"
        return os.path.isfile(self.archname)

    def extract(self):
        "Extract files from an archived submit dir"

        # Locate archive file
        if not self.is_archived():
            raise SubmitDirException("Submit dir not archived")

        # Update record in master db
        mdbsession = connection.connect_by_submitdir(
            self.submitdir, connection.DBType.MASTER
        )
        mdb = MasterDatabase(mdbsession)
        wf = mdb.get_master_workflow(self.wf_uuid)
        if wf is not None:
            wf.archived = False

        # Untar the files
        tar = tarfile.open(self.archname, "r:gz")
        tar.extractall(path=self.submitdir)
        tar.close()

        # Remove the tar file
        os.remove(self.archname)

        # Commit the workflow changes
        mdbsession.commit()
        mdbsession.close()

    def archive(self):
        "Archive a submit dir by adding files to a compressed archive"

        # Update record in master db
        mdbsession = connection.connect_by_submitdir(
            self.submitdir, connection.DBType.MASTER
        )
        mdb = MasterDatabase(mdbsession)
        wf = mdb.get_master_workflow(self.wf_uuid)
        if wf is not None:
            wf.archived = True

        # The set of files to exclude from the archive
        exclude = set()

        # Exclude braindump file
        exclude.add(self.braindump_file)

        # We use a temporary file so that we can determine if the archive step
        # completed successfully later
        tmparchname = os.path.join(self.submitdir, "archive.tmp.tar.gz")

        # We use a lock file to determine if cleanup is complete
        lockfile = os.path.join(self.submitdir, "archive.cleanup.lock")

        # If a previous archive was partially completed, then remove the
        # temporary file that was created
        if os.path.exists(tmparchname):
            os.unlink(tmparchname)

        # Exclude the temporary archive name so we don't add it to itself
        exclude.add(tmparchname)

        # We don't want the lock file to be saved, if it exists
        exclude.add(lockfile)

        # Also exclude the final archive name in case they try to run it again
        exclude.add(self.archname)

        # Ignore monitord files. This is needed so that tools like pegasus-statistics
        # will consider the workflow to be complete
        for name in ["monitord.started", "monitord.done", "monitord.log"]:
            exclude.add(os.path.join(self.submitdir, name))

        # Exclude stampede db
        for db in glob.glob(os.path.join(self.submitdir, "*.stampede.db")):
            exclude.add(db)

        # Exclude properties file
        for prop in glob.glob(os.path.join(self.submitdir, "pegasus.*.properties")):
            exclude.add(prop)

        # Visit all the files in the submit dir that we want to archive
        def visit(dirpath):
            for name in os.listdir(dirpath):
                filepath = os.path.join(dirpath, name)

                if filepath not in exclude:
                    yield name, filepath

        if self.is_archived() and not os.path.exists(lockfile):
            raise SubmitDirException("Submit directory already archived")

        if not self.is_archived():
            # Archive the files
            print("Creating archive...")
            tar = tarfile.open(name=tmparchname, mode="w:gz")
            for name, path in visit(self.submitdir):
                tar.add(name=path, arcname=name)
            tar.close()

            # This "commits" the archive step
            os.rename(tmparchname, self.archname)

        # Touch lockfile
        open(lockfile, "w").close()

        # Remove the files and directories
        # We do this here, instead of doing it in the loop above
        # because we want to make sure there are no errors in creating
        # the archive before we start removing files
        print("Removing files...")
        for name, path in visit(self.submitdir):
            if os.path.isfile(path) or os.path.islink(path):
                os.remove(path)
            else:
                shutil.rmtree(path)

        # This "commits" the file removal
        os.unlink(lockfile)

        # Commit the workflow changes
        mdbsession.commit()
        mdbsession.close()

    def move(self, dest):
        "Move this submit directory to dest"

        dest = os.path.abspath(dest)

        if os.path.isfile(dest):
            raise SubmitDirException("Destination is a file: %s" % dest)

        if os.path.isdir(dest):
            if os.path.exists(os.path.join(dest, "braindump.txt")):
                raise SubmitDirException("Destination is a submit dir: %s" % dest)
            dest = os.path.join(dest, os.path.basename(self.submitdir))

        # Verify that we aren't trying to move a subworkflow
        if self.is_subworkflow():
            raise SubmitDirException(
                "Subworkflows cannot be moved independent of the root workflow"
            )

        # Connect to master database
        mdbsession = connection.connect_by_submitdir(
            self.submitdir, connection.DBType.MASTER
        )
        mdb = MasterDatabase(mdbsession)

        # Get the workflow record from the master db
        db_url = None
        wf = mdb.get_master_workflow(self.wf_uuid)
        if wf is None:
            db_url = connection.url_by_submitdir(
                self.submitdir, connection.DBType.WORKFLOW
            )
        else:
            # We found an mdb record, so we need to update it

            # Save the master db's pointer
            db_url = wf.db_url

            # Update the master db's db_url
            # Note that this will only update the URL if it is an sqlite file
            # located in the submitdir
            log.info("Old master db_url: %s" % wf.db_url)
            wf.db_url = db_url.replace(self.submitdir, dest)
            log.info("New master db_url: %s" % wf.db_url)

            # Change the master db's submit_dir
            log.info("Old master submit_dir: %s" % wf.submit_dir)
            wf.submit_dir = dest
            log.info("New master submit_dir: %s" % wf.submit_dir)

        # Update the ensemble record if one exists
        ew = mdb.get_ensemble_workflow(self.wf_uuid)
        if ew is not None:
            log.info("Old ensemble submit dir: %s", ew.submitdir)
            ew.submitdir = dest
            log.info("New ensemble submit dir: %s", ew.submitdir)

        # Update the workflow database if we found one
        if db_url is not None:
            dbsession = connection.connect(db_url)
            db = WorkflowDatabase(dbsession)
            root_wf = db.get_workflow(self.wf_uuid)
            db.update_submit_dirs(root_wf.wf_id, self.submitdir, dest)
            dbsession.commit()
            dbsession.close()

        # Move all the files
        shutil.move(self.submitdir, dest)

        # Set new paths in the braindump file
        self.braindump["submit_dir"] = dest
        self.braindump["basedir"] = os.path.dirname(dest)
        utils.write_braindump(os.path.join(dest, "braindump.txt"), self.braindump)

        # Note that we do not need to update the properties file even though it
        # might contain DB URLs because it cannot contain a DB URL with the submit
        # dir in it.

        # TODO We might want to update all of the absolute paths in the condor submit files
        # if we plan on moving workflows that could be resubmitted in the future

        # TODO We might want to update the braindump files for subworkflows

        # Update master database
        mdbsession.commit()
        mdbsession.close()

        # Finally, update object
        self.submitdir = dest

    def delete(self):
        "Delete this submit dir and its entry in the master db"

        # Verify that we aren't trying to move a subworkflow
        if self.is_subworkflow():
            raise SubmitDirException(
                "Subworkflows cannot be deleted independent of the root workflow"
            )

        # Confirm that they want to delete the workflow
        while True:
            try:
                input = raw_input
            except NameError:
                pass
            answer = (
                input(
                    "Are you sure you want to delete this workflow? This operation cannot be undone. [y/n]: "
                )
                .strip()
                .lower()
            )
            if answer == "y":
                break
            if answer == "n":
                return

        # Connect to master database
        mdbsession = connection.connect_by_submitdir(
            self.submitdir, connection.DBType.MASTER
        )
        mdb = MasterDatabase(mdbsession)

        # Delete all of the records from the workflow db if they are not using
        # an sqlite db that is in the submit dir.
        db_url = connection.url_by_submitdir(self.submitdir, connection.DBType.WORKFLOW)
        if self.submitdir not in db_url:
            dbsession = connection.connect(db_url)
            db = WorkflowDatabase(dbsession)
            db.delete_workflow(self.wf_uuid)
            dbsession.commit()
            dbsession.close()

        # Delete the workflow
        mdb.delete_master_workflow(self.wf_uuid)

        # Remove all the files
        shutil.rmtree(self.submitdir)

        # Update master db
        mdbsession.commit()
        mdbsession.close()

    def attach(self):
        "Add a workflow to the master db"

        # Verify that we aren't trying to attach a subworkflow
        if self.is_subworkflow():
            raise SubmitDirException(
                "Subworkflows cannot be attached independent of the root workflow"
            )

        # Connect to master database
        mdbsession = connection.connect_by_submitdir(
            self.submitdir, connection.DBType.MASTER
        )
        mdb = MasterDatabase(mdbsession)

        # Check to see if it already exists and just update it
        wf = mdb.get_master_workflow(self.wf_uuid)
        if wf is not None:
            print("Workflow is already in master db")
            old_submit_dir = wf.submit_dir
            if old_submit_dir != self.submitdir:
                print("Updating path...")
                wf.submit_dir = self.submitdir
                wf.db_url = connection.url_by_submitdir(
                    self.submitdir, connection.DBType.WORKFLOW
                )
                mdbsession.commit()
            mdbsession.close()
            return

        # Connect to workflow db
        db_url = connection.url_by_submitdir(self.submitdir, connection.DBType.WORKFLOW)
        dbsession = connection.connect(db_url)
        db = WorkflowDatabase(dbsession)

        # Get workflow record
        wf = db.get_workflow(self.wf_uuid)
        if wf is None:
            print("No database record for that workflow exists")
            return

        # Update the workflow record
        wf.submit_dir = self.submitdir
        wf.db_url = db_url

        # Insert workflow record into master db
        mwf = MasterWorkflow()
        mwf.wf_uuid = wf.wf_uuid
        mwf.dax_label = wf.dax_label
        mwf.dax_version = wf.dax_version
        mwf.dax_file = wf.dax_file
        mwf.dag_file_name = wf.dag_file_name
        mwf.timestamp = wf.timestamp
        mwf.submit_hostname = wf.submit_hostname
        mwf.submit_dir = self.submitdir
        mwf.planner_arguments = wf.planner_arguments
        mwf.user = wf.user
        mwf.grid_dn = wf.grid_dn
        mwf.planner_version = wf.planner_version
        mwf.db_url = wf.db_url
        mwf.archived = self.is_archived()
        mdbsession.add(mwf)
        mdbsession.flush()  # We should have the new wf_id after this

        # Query states from workflow database
        states = db.get_workflow_states(wf.wf_id)

        # Insert states into master db
        for s in states:
            ms = MasterWorkflowstate()
            ms.wf_id = mwf.wf_id
            ms.state = s.state
            ms.timestamp = s.timestamp
            ms.restart_count = s.restart_count
            ms.status = s.status
            mdbsession.add(ms)
        mdbsession.flush()

        dbsession.commit()
        dbsession.close()

        mdbsession.commit()
        mdbsession.close()

    def detach(self, wf_uuid=None):
        "Remove any master db entries for the given root workflow"
        if self.submitdir_exists:
            # Verify that we aren't trying to detach a subworkflow
            if self.is_subworkflow():
                raise SubmitDirException(
                    "Subworkflows cannot be detached independent of the root workflow"
                )

            # Connect to master database
            mdbsession = connection.connect_by_submitdir(
                self.submitdir, connection.DBType.MASTER
            )
            mdb = MasterDatabase(mdbsession)

            # Check to see if it even exists
            wf = mdb.get_master_workflow(self.wf_uuid)
            if wf is None:
                print("Workflow is not in master DB")
            else:
                # Delete the workflow (this will delete the master_workflowstate entries as well)
                mdb.delete_master_workflow(self.wf_uuid)

            # Update the master db
            mdbsession.commit()
            mdbsession.close()

        else:
            # Connect to master database
            home = expanduser("~")
            mdbsession = connection.connect(
                "sqlite:///%s/.pegasus/workflow.db" % home,
                db_type=connection.DBType.MASTER,
            )
            mdb = MasterDatabase(mdbsession)

            try:
                if wf_uuid is None:
                    wfs = mdb.get_master_workflow_for_submitdir(self.submitdir)
                    if wfs:
                        msg = (
                            "Invalid submit dir: %s, Specify --wf-uuid <WF_UUID> to detach\n"
                            % self.submitdir
                        )
                        msg += (
                            "\tWorkflow UUID, DAX Label, Submit Hostname, Submit Dir.\n"
                        )
                        for wf in wfs:
                            msg += "\t{}, {}, {}, {}\n".format(
                                wf.wf_uuid,
                                wf.dax_label,
                                wf.submit_hostname,
                                wf.submit_dir,
                            )
                        raise SubmitDirException(msg)

                    else:
                        raise SubmitDirException(
                            "Invalid submit dir: %s" % self.submitdir
                        )

                else:
                    # Delete
                    mdb.delete_master_workflow(wf_uuid, submit_dir=self.submitdir)

                    # Update the master db
                    mdbsession.commit()

            finally:
                mdbsession.close()


class ExtractCommand(LoggingCommand):
    description = "Extract (uncompress) submit directory"
    usage = "Usage: %prog extract SUBMITDIR"

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify SUBMITDIR")

        SubmitDir(self.args[0]).extract()


class ArchiveCommand(LoggingCommand):
    description = "Archive (compress) submit directory"
    usage = "Usage: %prog archive SUBMITDIR"

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify SUBMITDIR")

        SubmitDir(self.args[0]).archive()


class MoveCommand(LoggingCommand):
    description = "Move a submit directory"
    usage = "Usage: %prog move SUBMITDIR DEST"

    def run(self):
        if len(self.args) != 2:
            self.parser.error("Specify SUBMITDIR and DEST")

        SubmitDir(self.args[0]).move(self.args[1])


class DeleteCommand(LoggingCommand):
    description = "Delete a submit directory and the associated DB entries"
    usage = "Usage: %prog delete SUBMITDIR"

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify SUBMITDIR")

        SubmitDir(self.args[0]).delete()


class AttachCommand(LoggingCommand):
    description = "Attach a submit dir to the master db (dashboard)"
    usage = "Usage: %prog attach SUBMITDIR"

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify SUBMITDIR")

        SubmitDir(self.args[0]).attach()


class DetachCommand(LoggingCommand):
    description = "Detach a submit dir from the master db (dashboard)"
    usage = "Usage: %prog detach SUBMITDIR"

    def __init__(self):
        LoggingCommand.__init__(self)
        self.parser.add_option(
            "-i",
            "--wf-uuid",
            dest="wf_uuid",
            help="Specify wf_uuid of the workflow to be detached.",
        )

    def run(self):
        if len(self.args) != 1:
            self.parser.error("Specify SUBMITDIR")

        wf_uuid = self.options.wf_uuid
        SubmitDir(self.args[0], raise_err=False).detach(wf_uuid=wf_uuid)


class SubmitDirCommand(CompoundCommand):
    description = "Manages submit directories"
    commands = [
        ("archive", ArchiveCommand),
        ("extract", ExtractCommand),
        ("move", MoveCommand),
        ("delete", DeleteCommand),
        ("attach", AttachCommand),
        ("detach", DetachCommand),
    ]
    aliases = {
        "ar": "archive",
        "ex": "extract",
        "mv": "move",
        "rm": "delete",
        "at": "attach",
        "dt": "detach",
    }


def main():
    "The entry point for pegasus-submitdir"
    SubmitDirCommand().main()



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__author__ = "Rafael Ferreira da Silva"

DB_VERSION = 0

import logging

from sqlalchemy.exc import *

from Pegasus.db.admin.admin_loader import *
from Pegasus.db.admin.versions.base_version import *
from Pegasus.db.schema import *

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        log.info("Updating to version %s" % DB_VERSION)
        query = "ALTER TABLE invocation ADD COLUMN remote_cpu_time NUMERIC(10,3) NULL"
        if self.db.connection().dialect.name != "sqlite":
            query += " AFTER remote_duration"
        self.execute(query)
        self.execute(
            "ALTER TABLE job_instance ADD COLUMN multiplier_factor INT NOT NULL DEFAULT 1"
        )
        self.execute("ALTER TABLE job_instance ADD COLUMN exitcode INT NULL")

        success = ["JOB_SUCCESS", "POST_SCRIPT_SUCCESS"]
        failure = [
            "PRE_SCRIPT_FAILED",
            "SUBMIT_FAILED",
            "JOB_FAILURE",
            "POST_SCRIPT_FAILED",
        ]

        try:
            q = self.db.query(JobInstance.job_instance_id).order_by(
                JobInstance.job_instance_id
            )
            for r in q.all():
                qq = self.db.query(Jobstate.state)
                qq = qq.filter(Jobstate.job_instance_id == r.job_instance_id)
                qq = qq.order_by(Jobstate.jobstate_submit_seq.desc()).limit(1)
                for rr in qq.all():
                    if rr.state in success:
                        self.db.execute(
                            "UPDATE job_instance SET exitcode = 0 WHERE job_instance_id = %s"
                            % r.job_instance_id
                        )
                    elif rr.state in failure:
                        self.db.execute(
                            "UPDATE job_instance SET exitcode = 256 WHERE job_instance_id = %s"
                            % r.job_instance_id
                        )
                    else:
                        pass
        except (OperationalError, ProgrammingError):
            pass

        self.db.commit()

    def downgrade(self, force=False):
        """ Downgrade to this version will not be allowed."""

    def execute(self, query):
        try:
            self.db.execute(query)
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)
